# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# yapf: disable
# isort: skip_file

import re
from itertools import islice
from typing import List, Callable, Optional, Tuple

import jax
import jax.numpy as jnp
import torch
from flax import nnx
from flax.typing import PRNGKey
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P
from vllm.config import VllmConfig

from tpu_inference.distributed.jax_parallel_state import get_pp_group
from tpu_inference.layers.jax.attention.attention import AttentionMetadata
from tpu_inference.layers.jax.attention.llama4_attention import Llama4Attention
from tpu_inference.layers.jax.constants import KVCacheType
from tpu_inference.layers.jax.layers import DenseFFW, Embedder, LMhead, RMSNorm
from tpu_inference.layers.jax.rope import \
    Llama4VisionRotaryEmbedding
from tpu_inference.layers.jax.misc import shard_put
from tpu_inference.layers.jax.pp_utils import (PPMissingLayer,
                                               get_start_end_layer)
from tpu_inference.layers.jax.transformer_block import TransformerBlock
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.llama4 import (JAXLlama4MultiModalProjector,
                                             JAXLlama4VisionModel)
from tpu_inference.models.jax.utils.multi_modal_utils import \
    merge_multimodal_embeddings
from tpu_inference.models.jax.jax_intermediate_tensor import \
    JaxIntermediateTensors
from tpu_inference.models.jax.utils.weight_utils import (
    BaseWeightLoader, _is_pp_missing_layer, get_param, print_param_info,
    reshape_params, transpose_params)

import torchax

logger = init_logger(__name__)


class LlamaGuard4WeightLoader(BaseWeightLoader):

    def __init__(self, vllm_config: VllmConfig, hidden_size, attn_heads,
                 num_key_value_heads, attn_head_dim):
        super().__init__(
            vllm_config,
            framework="flax",
            filter_regex=
            r"^(language_model|vision_model|multi_modal_projector)\..*",
        )
        # Set is_runai_streamer so we can use this in load_weights
        self.is_runai_streamer = getattr(
            getattr(vllm_config, 'load_config', None), 'load_format',
            None) == 'runai_streamer'
        self.is_verbose = getattr(vllm_config.additional_config, "is_verbose",
                                  False)

        # --- Language model transpose map ---
        self._language_transpose_map = {
            "self_attn.q_proj": (2, 0, 1),
            "self_attn.k_proj": (2, 0, 1),
            "self_attn.v_proj": (2, 0, 1),
            "self_attn.o_proj": (1, 2, 0),
            "lm_head": (1, 0),
            "feed_forward.down_proj": (1, 0),
            "feed_forward.gate_proj": (1, 0),
            "feed_forward.up_proj": (1, 0),
        }

        # Vision model transpose map
        self._vision_transpose_map = {
            "patch_embedding.linear":
            (2, 3, 1, 0),
            "self_attn.q_proj": (2, 0, 1),
            "self_attn.k_proj": (2, 0, 1),
            "self_attn.v_proj": (2, 0, 1),
            "self_attn.o_proj": (1, 2, 0),
            "mlp.fc1": (1, 0),
            "mlp.fc2": (1, 0),
            "vision_adapter.mlp.fc1": (1, 0),
            "vision_adapter.mlp.fc2": (1, 0),
        }

        # --- Projector transpose map ---
        self._projector_transpose_map = {
            "linear_1": (1, 0),
        }

        # --- Shape Maps ---
        self._weight_shape_map = {
            "q_proj": (attn_heads, attn_head_dim, hidden_size),
            "k_proj": (num_key_value_heads, attn_head_dim, hidden_size),
            "v_proj": (num_key_value_heads, attn_head_dim, hidden_size),
            "o_proj": (hidden_size, attn_heads, attn_head_dim),
        }

        vision_config = vllm_config.model_config.hf_config.vision_config
        vision_hidden_size = vision_config.hidden_size
        vision_attn_heads = vision_config.num_attention_heads
        vision_head_dim = vision_hidden_size // vision_attn_heads

        self._vision_weight_shape_map = {
            "patch_embedding.linear":
            (vision_hidden_size, 3, vision_config.patch_size,
             vision_config.patch_size),
            "q_proj": (vision_attn_heads, vision_head_dim, vision_hidden_size),
            "k_proj": (vision_attn_heads, vision_head_dim, vision_hidden_size),
            "v_proj": (vision_attn_heads, vision_head_dim, vision_hidden_size),
            "o_proj": (vision_hidden_size, vision_attn_heads, vision_head_dim),
        }

        self._loaded_to_standardized_keys = {
            # --- Text Model Mappings ---
            "language_model.model.embed_tokens.weight":
            "embedder.input_embedding_table_VD",
            "language_model.lm_head.weight":
            "lm_head.input_embedding_table_DV",
            "language_model.model.norm.weight":
            "final_norm.scale",
            "language_model.model.layers.*.input_layernorm.weight":
            "layers.*.pre_attention_norm.scale",
            "language_model.model.layers.*.post_attention_layernorm.weight":
            "layers.*.pre_mlp_norm.scale",
            "language_model.model.layers.*.self_attn.q_proj.weight":
            "layers.*.attn.kernel_q_proj_DNH",
            "language_model.model.layers.*.self_attn.k_proj.weight":
            "layers.*.attn.kernel_k_proj_DKH",
            "language_model.model.layers.*.self_attn.v_proj.weight":
            "layers.*.attn.kernel_v_proj_DKH",
            "language_model.model.layers.*.self_attn.o_proj.weight":
            "layers.*.attn.kernel_o_proj_NHD",
            "language_model.model.layers.*.feed_forward.gate_proj.weight":
            "layers.*.custom_module.kernel_gating_DF",
            "language_model.model.layers.*.feed_forward.up_proj.weight":
            "layers.*.custom_module.kernel_up_proj_DF",
            "language_model.model.layers.*.feed_forward.down_proj.weight":
            "layers.*.custom_module.kernel_down_proj_FD",

            # --- Vision Model Mappings ---
            "vision_model.patch_embedding.linear.weight":
            "vision_model.patch_embedding.linear.kernel",
            "vision_model.class_embedding":
            "vision_model.class_embedding",
            "vision_model.positional_embedding_vlm":
            "vision_model.positional_embedding_vlm",
            "vision_model.layernorm_pre.weight":
            "vision_model.layernorm_pre.scale",
            "vision_model.layernorm_pre.bias":
            "vision_model.layernorm_pre.bias",
            "vision_model.layernorm_post.weight":
            "vision_model.layernorm_post.scale",
            "vision_model.layernorm_post.bias":
            "vision_model.layernorm_post.bias",

            # Vision Encoder Layer Weights
            "vision_model.model.layers.*.input_layernorm.weight":
            "vision_model.model.layers.*.input_layernorm.scale",
            "vision_model.model.layers.*.input_layernorm.bias":
            "vision_model.model.layers.*.input_layernorm.bias",
            "vision_model.model.layers.*.post_attention_layernorm.weight":
            "vision_model.model.layers.*.post_attention_layernorm.scale",
            "vision_model.model.layers.*.post_attention_layernorm.bias":
            "vision_model.model.layers.*.post_attention_layernorm.bias",

            # ATTENTION KERNELS
            "vision_model.model.layers.*.self_attn.q_proj.weight":
            "vision_model.model.layers.*.self_attn.kernel_q_proj_DNH",
            "vision_model.model.layers.*.self_attn.k_proj.weight":
            "vision_model.model.layers.*.self_attn.kernel_k_proj_DKH",
            "vision_model.model.layers.*.self_attn.v_proj.weight":
            "vision_model.model.layers.*.self_attn.kernel_v_proj_DKH",
            "vision_model.model.layers.*.self_attn.o_proj.weight":
            "vision_model.model.layers.*.self_attn.kernel_o_proj_NHD",
            "vision_model.model.layers.*.self_attn.q_proj.bias":
            "vision_model.model.layers.*.self_attn.bias_q_proj_NH",
            "vision_model.model.layers.*.self_attn.k_proj.bias":
            "vision_model.model.layers.*.self_attn.bias_k_proj_KH",
            "vision_model.model.layers.*.self_attn.v_proj.bias":
            "vision_model.model.layers.*.self_attn.bias_v_proj_KH",
            "vision_model.model.layers.*.self_attn.o_proj.bias":
            "vision_model.model.layers.*.self_attn.bias_o_proj_D",

            # VISION MLP WEIGHTS (FC1/FC2)
            "vision_model.model.layers.*.mlp.fc1.weight":
            "vision_model.model.layers.*.mlp.fc1.kernel",
            "vision_model.model.layers.*.mlp.fc1.bias":
            "vision_model.model.layers.*.mlp.fc1.bias",
            "vision_model.model.layers.*.mlp.fc2.weight":
            "vision_model.model.layers.*.mlp.fc2.kernel",
            "vision_model.model.layers.*.mlp.fc2.bias":
            "vision_model.model.layers.*.mlp.fc2.bias",

            # Vision Adapter (Pixel Shuffle MLP)
            "vision_model.vision_adapter.mlp.fc1.weight":
            "vision_model.vision_adapter.pixel_shuffle_mlp.fc1.kernel",
            "vision_model.vision_adapter.mlp.fc2.weight":
            "vision_model.vision_adapter.pixel_shuffle_mlp.fc2.kernel",

            # Multimodal Projector
            "multi_modal_projector.linear_1.weight":
            "multi_modal_projector.linear.kernel",
        }

    def map_loaded_to_standardized_name(self, loaded_key: str) -> str:

        # 1. Check if the key contains the layer pattern
        layer_match = re.search(r"layers\.(\d+)", loaded_key)

        if layer_match:
            # If it's a layer weight: extract number and map with wildcard
            layer_num = layer_match.group(1)
            layer_key = re.sub(r"layers\.\d+", "layers.*", loaded_key)

            # Map the wildcard key to the standardized path
            mapped_key = self._loaded_to_standardized_keys.get(
                layer_key, loaded_key)

            # Substitute the wildcard with the actual layer number
            mapped_key = re.sub(r"layers\.\*", f"layers.{layer_num}",
                                mapped_key)

        else:
            # 2. If it's a non-layer weight (lm_head, embed_tokens, etc.): map directly
            mapped_key = self._loaded_to_standardized_keys.get(
                loaded_key, loaded_key)

        return mapped_key

    def load_weights(self, model_for_loading: nnx.Module):
        model_params = nnx.state(model_for_loading)

        self.pp_missing_layers = []
        for path, module in nnx.iter_graph(model_for_loading):
            if isinstance(module, PPMissingLayer):
                layer_name = ".".join([str(s) for s in path])
                self.pp_missing_layers.append(layer_name)

        with jax.default_device(jax.devices("cpu")[0]):
            for loaded_name, loaded_weight in self.get_weights_iterator():
                if self.is_runai_streamer:
                    if torchax is not None:
                        env = torchax.default_env()
                        loaded_weight = env.t2j_copy(loaded_weight)
                    else:
                        loaded_weight = jnp.array(loaded_weight.cpu().numpy())

                mapped_name = self.map_loaded_to_standardized_name(loaded_name)

                if _is_pp_missing_layer(mapped_name, self.pp_missing_layers):
                    continue

                try:
                    model_weight = get_param(model_params, mapped_name)
                except KeyError:
                    # Optional debug logging for skipped weights
                    if self.is_verbose:
                        print(f"Skipping {loaded_name}")
                    continue

                # --- Vision Model Bias Reshaping ---
                if "vision_model" in loaded_name and ".self_attn." in loaded_name and loaded_name.endswith(
                        ".bias"):
                    vision_config = model_for_loading.vision_config
                    vision_heads = vision_config.num_attention_heads
                    vision_head_dim = vision_config.hidden_size // vision_heads
                    if "q_proj.bias" in loaded_name or "k_proj.bias" in loaded_name or "v_proj.bias" in loaded_name:
                        loaded_weight = jnp.reshape(
                            loaded_weight, (vision_heads, vision_head_dim))

                # --- General Weight Reshaping and Transposition ---
                if not loaded_name.endswith(".bias"):
                    shape_map_to_use = None
                    transpose_map_to_use = None

                    if "language_model" in loaded_name:
                        shape_map_to_use = self._weight_shape_map
                        transpose_map_to_use = self._language_transpose_map
                    elif "vision_model" in loaded_name:
                        shape_map_to_use = self._vision_weight_shape_map
                        transpose_map_to_use = self._vision_transpose_map
                    elif "multi_modal_projector" in loaded_name:
                        transpose_map_to_use = self._projector_transpose_map

                    if shape_map_to_use:
                        loaded_weight = reshape_params(loaded_name,
                                                       loaded_weight,
                                                       shape_map_to_use)

                    if transpose_map_to_use:
                        loaded_weight = transpose_params(loaded_name,
                                                         loaded_weight,
                                                         transpose_map_to_use)

                    if "patch_embedding.linear" in mapped_name:
                        loaded_weight = loaded_weight.reshape(
                            -1, loaded_weight.shape[-1])

                if model_weight.value.shape != loaded_weight.shape:
                    raise ValueError(
                        f"Loaded shape for {loaded_name}: {loaded_weight.shape} "
                        f"does not match model shape for {mapped_name}: {model_weight.value.shape}!"
                    )

                if self.is_verbose:
                    print_param_info(model_weight, loaded_name)

                model_weight.value = shard_put(loaded_weight,
                                               model_weight.out_sharding,
                                               mesh=model_for_loading.mesh)

        nnx.update(model_for_loading, model_params)


class LlamaGuard4ForCausalLM(nnx.Module):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: PRNGKey,
                 mesh: Mesh,
                 force_random_weights: bool = False):
        assert mesh is not None

        self.vllm_config = vllm_config
        self.vllm_config.model_config.dtype = torch.bfloat16
        self.model_config = vllm_config.model_config
        self.text_config = self.model_config.hf_config.text_config
        self.vision_config = self.model_config.hf_config.vision_config

        self.projector_config_dict = {
            'vision_config': self.vision_config,
            'text_config': self.text_config,
        }

        self.mesh = mesh
        self.is_verbose = getattr(self.vllm_config.additional_config,
                                   "is_verbose", False)

        self.use_qk_norm = getattr(self.text_config, "use_qk_norm", True)

        vocab_size = self.model_config.get_vocab_size()
        self.hidden_size = self.model_config.get_hidden_size()

        self.dtype: jnp.dtype = jnp.bfloat16

        self.num_layers: int = getattr(self.text_config, "num_layers", 48)
        hidden_act: str = getattr(self.text_config, "hidden_act", "silu")

        rms_norm_eps = getattr(self.text_config, "rms_norm_eps", 1e-5)
        self.num_attention_heads = getattr(self.text_config,
                                           "num_attention_heads", 40)
        self.num_key_value_heads = getattr(self.text_config,
                                           "num_key_value_heads", 8)
        self.head_dim = getattr(self.text_config, "head_dim", 128)

        intermediate_size = getattr(self.text_config, "intermediate_size",
                                    8192)

        self.rope_theta_text = getattr(self.text_config, "rope_theta",
                                       500000.0)
        self.rope_scaling = getattr(self.text_config, "rope_scaling")

        self.image_token_id = getattr(self.model_config.hf_config,
                                      "image_token_index")

        self.rng = nnx.Rngs(rng)

        self.is_first_rank = get_pp_group().is_first_rank
        self.is_last_rank = get_pp_group().is_last_rank

        if self.is_first_rank or (self.model_config.hf_config.tie_word_embeddings
                                  and self.is_last_rank):
            self.embedder = Embedder(
                vocab_size=vocab_size,
                hidden_size=self.hidden_size,
                dtype=self.dtype,
                vd_sharding=P(('data', 'model'), None),
                rngs=self.rng,
                random_init=force_random_weights,
            )
        else:
            self.embedder = PPMissingLayer()

        self.vision_rope = Llama4VisionRotaryEmbedding(
            image_size=self.vision_config.image_size,
            patch_size=self.vision_config.patch_size,
            hidden_size=self.vision_config.hidden_size,
            num_attention_heads=self.vision_config.num_attention_heads,
            rope_theta=self.vision_config.rope_theta,
            dtype=self.dtype,
        )

        self.vision_model = JAXLlama4VisionModel(
            self.vision_config,
            rngs=self.rng,
            mesh=self.mesh,
            dtype=self.dtype,
            random_init=force_random_weights,
            vision_rope=self.vision_rope)


        self.multi_modal_projector = JAXLlama4MultiModalProjector(
            self.projector_config_dict,
            rngs=self.rng,
            dtype=self.dtype,
            random_init=force_random_weights,
        )

        self.layers = nnx.List([])
        self.start_layer, self.end_layer = get_start_end_layer(
            self.num_layers,
            get_pp_group().rank_in_group,
            get_pp_group().world_size)

        for i in range(self.start_layer):
            self.layers.append(PPMissingLayer())

        for i in range(self.start_layer, self.end_layer):
            use_attention_rope = True

            custom_module = DenseFFW(dtype=self.dtype,
                                     hidden_act=hidden_act,
                                     hidden_size=self.hidden_size,
                                     intermediate_size=intermediate_size,
                                     random_init=force_random_weights,
                                     rngs=self.rng,
                                     df_sharding=P(None, 'model'),
                                     fd_sharding=P('model', None),
                                     activation_ffw_td=P('data', None),
                                     mesh=self.mesh)

            attn = Llama4Attention(
                hidden_size=self.hidden_size,
                dtype=self.dtype,
                num_attention_heads=self.num_attention_heads,
                num_key_value_heads=self.num_key_value_heads,
                head_dim=self.head_dim,
                rope_theta=self.rope_theta_text,
                rope_scaling={
                    "scale_factor":
                    self.rope_scaling["factor"],
                    "low_freq_factor":
                    self.rope_scaling["low_freq_factor"],
                    "high_freq_factor":
                    self.rope_scaling["high_freq_factor"],
                    "original_max_position_embeddings":
                    self.rope_scaling["original_max_position_embeddings"]
                },
                rngs=self.rng,
                rope_input_ordering="interleaved",
                # TODO (jacobplatin): we should refactor this to pass a dtype (or config) directly
                kv_cache_dtype=vllm_config.cache_config.cache_dtype,
                temperature_tuning=True,
                temperature_tuning_scale=0.1,
                temperature_tuning_floor_scale=8192,
                use_qk_norm=self.use_qk_norm,
                attention_chunk_size=None if use_attention_rope else 8192,
                mesh=self.mesh,
                random_init=force_random_weights,
                activation_attention_td=P('data', 'model'),
                activation_q_td=P('data', 'model'),
                query_tnh=P('data', 'model', None),
                keyvalue_skh=P('data', 'model', None),
                activation_attention_out_td=P('data', 'model'),
                attn_o_tnh=P('data', 'model', None),
                dnh_sharding=(None, 'model', None),
                dkh_sharding=(None, 'model', None),
                nhd_sharding=('model', None, None),
            )

            pre_attention_norm = RMSNorm(
                dims=self.hidden_size,
                random_init=force_random_weights,
                epsilon=rms_norm_eps,
                rngs=self.rng,
                activation_ffw_td=P('data', None),
                with_scale=True,
                dtype=self.dtype,
            )

            pre_mlp_norm = RMSNorm(
                dims=self.hidden_size,
                activation_ffw_td=P('data', None),
                epsilon=rms_norm_eps,
                rngs=self.rng,
                with_scale=True,
                dtype=self.dtype,
                random_init=force_random_weights,
            )

            block = TransformerBlock(custom_module=custom_module,
                                     attn=attn,
                                     pre_attention_norm=pre_attention_norm,
                                     pre_mlp_norm=pre_mlp_norm,
                                     use_attention_rope=use_attention_rope)
            self.layers.append(block)

        for i in range(self.end_layer, self.num_layers):
            self.layers.append(PPMissingLayer())

        if self.is_last_rank:
            self.final_norm = RMSNorm(
                dims=self.hidden_size,
                activation_ffw_td=P(),
                epsilon=rms_norm_eps,
                rngs=self.rng,
                with_scale=True,
                dtype=self.dtype,
                random_init=force_random_weights,
            )
            self.lm_head = LMhead(vocab_size=vocab_size,
                                  hidden_size=self.hidden_size,
                                  dtype=self.dtype,
                                  rngs=self.rng,
                                  dv_sharding=P(None, ('data', 'model')),
                                  random_init=force_random_weights)
        else:
            self.final_norm = PPMissingLayer()
            self.lm_head = PPMissingLayer()

        if self.is_verbose:
            self._print_model_architecture()

    def _print_model_architecture(self):
        logger.info("### Embedding ###")
        nnx.display(self.embedder)
        logger.info("\n### Layers ###")
        for i, layer in enumerate(self.layers):
            logger.info(f"\n--- Layer {i} ---")
            nnx.display(layer)
        logger.info("\n### LM Head ###")
        nnx.display(self.lm_head)

    # In LlamaGuard4ForCausalLM
    @nnx.jit
    def _compute_vision_features_jit(self, pixel_values: jax.Array) -> jax.Array:
        vision_outputs = self.vision_model(pixel_values)
        projected = self.multi_modal_projector(vision_outputs)
        return projected

    def precompile_vision_encoder(self, run_compilation_fn: Callable) -> None:
        warmup_config = self.vllm_config.additional_config.get(
            "vision_warmup_config")
        if not warmup_config:
            return

        image_shapes = warmup_config.get("image_shapes", [])

        for (h, w) in image_shapes:
            # Create dummy input: Batch=1, H, W, Channels=3
            # We must use the same layout (NHWC) as used in execution
            dummy_pixel_values = jnp.ones((1, h, w, 3), dtype=jnp.bfloat16)

            run_compilation_fn(
                "vision_encoder_projector",
                self._compute_vision_features_jit,
                dummy_pixel_values
            )

    def load_weights(self, rng: jax.Array, cache_dir: Optional[str] = None):
        self.rng = nnx.Rngs(rng)

        weight_loader = LlamaGuard4WeightLoader(
            vllm_config=self.vllm_config,
            hidden_size=self.hidden_size,
            attn_heads=self.num_attention_heads,
            num_key_value_heads=self.num_key_value_heads,
            attn_head_dim=self.head_dim)
        weight_loader.load_weights(self)

        logger.info("Materializing Llama 4 Vision RoPE frequencies...")
        self.vision_rope.__post_init__()

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
        *args,
    ) -> Tuple[List[KVCacheType], jax.Array]:
        is_prefill = False

        if self.is_first_rank:
            if inputs_embeds is not None:
                x_TD = inputs_embeds
            elif input_ids is not None:
                x_TD = self.embedder.encode(input_ids)
            else:
                raise ValueError(
                    "Cannot run forward pass: Both input_ids and inputs_embeds are None."
                )
        else:
            # For pipeline parallelism (not active here, but good practice)
            # x_TD would come from intermediate_tensors
            pass

        for (i, block) in enumerate(
                islice(self.layers, self.start_layer, self.end_layer)):
            kv_cache = kv_caches[i]
            new_kv_cache, x_TD = block(x_TD, is_prefill, kv_cache,
                                       attention_metadata)
            jax.block_until_ready(x_TD)
            kv_caches[i] = new_kv_cache

        if not self.is_last_rank:
            return kv_caches, JaxIntermediateTensors({"hidden_states": x_TD}), []

        final_activation_TD = self.final_norm(x_TD)
        return kv_caches, final_activation_TD, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        logits_TV = jnp.dot(hidden_states,
                            self.lm_head.input_embedding_table_DV.value)
        return logits_TV

    def embed_input_ids(
            self,
            input_ids: jax.Array,
            multimodal_embeddings: jax.Array | None = None,
            *,
            is_multimodal: jax.Array | None = None,
        ) -> jax.Array:

        del is_multimodal


        inputs_embeds = self.embedder.encode(input_ids)


        if multimodal_embeddings is not None and multimodal_embeddings.shape[
                0] > 0:
            inputs_embeds = merge_multimodal_embeddings(
                input_ids, inputs_embeds, multimodal_embeddings,
                [self.image_token_id])

        return inputs_embeds

    def embed_multimodal(self, image_grid_thw, **kwargs) -> List[jax.Array]:
        pixel_values = kwargs.pop("pixel_values")
        patches_per_image = kwargs.pop("patches_per_image", None)

        if pixel_values is None:
            return []

        # Ensure JAX handling
        pixel_values = jnp.asarray(pixel_values, dtype=jnp.bfloat16)

        # 1. Transpose Input from NCHW to NHWC
        # vLLM/HF loaders typically give NCHW (Channels First)
        # We ensure NHWC (Channels Last) for the JAX Vision Model
        if pixel_values.ndim == 4 and pixel_values.shape[1] == 3:
            pixel_values = jnp.transpose(pixel_values, (0, 2, 3, 1))

        # 2. Run Vision Encoder (on NHWC data)
        projected_vision_features = self.multi_modal_projector(
            self.vision_model(pixel_values))

        tokens_per_tile = projected_vision_features.shape[1]
        hidden_dim = projected_vision_features.shape[2]

        # 3. Flatten tokens
        all_tokens_flat = projected_vision_features.reshape(-1, hidden_dim)

        # 4. Split batch logic
        if hasattr(patches_per_image, 'tolist'):
            tile_counts = patches_per_image.tolist()
        else:
            tile_counts = list(patches_per_image)
        if isinstance(tile_counts, list) and isinstance(tile_counts[0], list):
            import itertools
            tile_counts = list(itertools.chain.from_iterable(tile_counts))
        split_sizes = [c * tokens_per_tile for c in tile_counts]

        if not split_sizes or sum(split_sizes) == 0:
            return []

        if sum(split_sizes) != all_tokens_flat.shape[0]:
            raise ValueError(
                "Sum of split sizes must match the total number of tokens.")

        split_indices = jnp.cumsum(jnp.array(split_sizes[:-1]))
        output_list = jnp.split(all_tokens_flat, split_indices)

        return list(output_list)


# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict, List, Optional

import jax
from jax.sharding import Mesh


@dataclass
class VllmModelWrapperContext:
    kv_caches: List[jax.Array]
    mesh: Mesh
    layer_name_to_kvcache_index: Dict[str, int]


_vllm_model_wrapper_context: Optional[VllmModelWrapperContext] = None


def get_vllm_model_wrapper_context() -> VllmModelWrapperContext:
    assert _vllm_model_wrapper_context is not None, (
        "VllmModelWrapperContext is not set. "
        "Please use `set_vllm_model_wrapper_context` to set the VllmModelWrapperContext."
    )
    return _vllm_model_wrapper_context


@contextmanager
def set_vllm_model_wrapper_context(
    *,
    kv_caches: List[jax.Array],
    mesh: Mesh,
    layer_name_to_kvcache_index: Dict[str, int] = None,
):
    global _vllm_model_wrapper_context
    prev_context = _vllm_model_wrapper_context
    _vllm_model_wrapper_context = VllmModelWrapperContext(
        kv_caches=kv_caches,
        mesh=mesh,
        layer_name_to_kvcache_index=layer_name_to_kvcache_index,
    )

    try:
        yield
    finally:
        _vllm_model_wrapper_context = prev_context



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from flax import nnx
from flax.typing import PRNGKey
from jax.sharding import Mesh
from transformers.models.qwen2_5_vl.configuration_qwen2_5_vl import \
    Qwen2_5_VLConfig
from vllm.config import (CacheConfig, DeviceConfig, MultiModalConfig,
                         ParallelConfig, SchedulerConfig)

from tpu_inference.distributed.jax_parallel_state import \
    init_pp_distributed_environment
from tpu_inference.layers.jax.pp_utils import PPMissingLayer
from tpu_inference.models.jax.jax_intermediate_tensor import \
    JaxIntermediateTensors
# Import the module itself to allow patching
# Corrected imports for the code under test
from tpu_inference.models.jax.qwen2_5_vl import (
    AttentionMetadata, Qwen2_5_VisionAttention, Qwen2_5_VisionBlock,
    Qwen2_5_VisionMLP, Qwen2_5_VisionPatchEmbed, Qwen2_5_VisionPatchMerger,
    Qwen2_5_VisionRotaryEmbedding, Qwen2_5_VisionTransformer,
    Qwen2_5_VLForConditionalGeneration, Qwen2_5_VLImagePixelInputs, SegmentIds,
    apply_rotary_pos_emb_vision, generate_window_segment_ids)


# --- Configuration Mocking ---
class MockModelConfig:

    def __init__(self, hf_config, dtype):
        self.hf_config = hf_config
        self.dtype = dtype
        self.multimodal_config = MultiModalConfig()
        self.model = "mock_qwen2_5_vl"
        # Add other attributes if needed by the code
        self.tokenizer = "mock_tokenizer"
        self.tokenizer_mode = "auto"
        self.trust_remote_code = True
        self.seed = 0

    def is_multimodal_model(self):
        return True

    def get_hidden_size(self):
        return self.hf_config.hidden_size

    def get_head_size(self):
        return self.hf_config.hidden_size // self.hf_config.num_attention_heads

    def get_vocab_size(self):
        return self.hf_config.vocab_size


class MockVllmConfig:
    """A mock VllmConfig sufficient for testing the Qwen2.5 VL model."""

    def __init__(self, tie_word_embeddings: bool = False):
        vision_config = {
            "hidden_size": 16,
            "intermediate_size": 32,
            "patch_size": 14,
            "image_size": 28,
            "temporal_patch_size": 2,
            "in_channels": 3,
            "window_size": 28,
            "spatial_merge_size": 2,
            "fullatt_block_indexes": [0],
            "out_hidden_size": 24,
            "depth": 2,
            "hidden_act": "gelu",
            "num_heads": 2,
        }
        hf_config = Qwen2_5_VLConfig(
            vision_config=vision_config,
            hidden_size=16,
            num_hidden_layers=2,
            num_attention_heads=2,
            num_key_value_heads=2,
            intermediate_size=32,
            rms_norm_eps=1e-6,
            image_token_id=200000,
            video_token_id=200001,
            tie_word_embeddings=tie_word_embeddings,
            vocab_size=32000,
            rope_theta=1000000.0,
        )
        self.model_config = MockModelConfig(hf_config, jnp.bfloat16)
        self.cache_config = MagicMock(spec=CacheConfig)
        self.cache_config.cache_dtype = "auto"
        self.parallelism_config = MagicMock(spec=ParallelConfig)
        self.scheduler_config = MagicMock(spec=SchedulerConfig)
        self.device_config = MagicMock(spec=DeviceConfig)
        self.load_config = MagicMock()
        self.extra_configs = {}
        self.additional_config = {}
        self.quant_config = None


@pytest.fixture(scope="module")
def mesh():
    """Creates a mesh with all required axes for testing."""
    if not jax.devices():
        pytest.skip("No JAX devices available for mesh creation.")
    devices = np.array(jax.local_devices())
    return Mesh(devices.reshape((len(devices), 1, 1)),
                axis_names=('data', 'attn_dp', 'model'))


@pytest.fixture
def rng() -> PRNGKey:
    """Provides a reusable JAX PRNGKey."""
    return jax.random.PRNGKey(42)


@pytest.fixture
def mock_vllm_config() -> MockVllmConfig:
    return MockVllmConfig()


@pytest.fixture
def rngs(rng: PRNGKey) -> nnx.Rngs:
    return nnx.Rngs(params=rng)


@pytest.fixture(autouse=True, scope="module")
def initialize_pp():
    init_pp_distributed_environment(
        ip="",
        rank=0,
        world_size=1,
        device=jax.devices()[0],
        need_pp=False,
    )


# --- Test Classes ---
class TestUtils:

    def test_apply_rotary_pos_emb_vision(self, rng: PRNGKey):
        B, T, N, H = 1, 10, 2, 8
        x = jax.random.normal(rng, (B, T, N, H))
        rotary_pos_emb = jax.random.normal(rng, (T, H // 2))
        x_rotated = apply_rotary_pos_emb_vision(x, rotary_pos_emb)
        assert x_rotated.shape == (B, T, N, H)

    def test_generate_window_segment_ids(self):
        cu_seqlens = jnp.array([0, 5, 10])
        seq_len = 10
        padded_seq_len = 16
        segment_ids = generate_window_segment_ids(cu_seqlens, seq_len,
                                                  padded_seq_len)
        assert isinstance(segment_ids, SegmentIds)
        assert segment_ids.q.shape == (1, padded_seq_len)
        assert segment_ids.kv.shape == (1, padded_seq_len)
        expected_q = np.array(
            [[1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0]])
        np.testing.assert_array_equal(segment_ids.q, expected_q)
        np.testing.assert_array_equal(segment_ids.kv, expected_q)


class TestQwen2_5_VisionMLP:

    def test_forward(
        self,
        mock_vllm_config: MockVllmConfig,
        rngs: nnx.Rngs,
        mesh: Mesh,
    ):
        config = mock_vllm_config.model_config.hf_config.vision_config
        dtype = mock_vllm_config.model_config.dtype
        with jax.set_mesh(mesh):
            mlp = Qwen2_5_VisionMLP(config, dtype, rngs)
        x = jnp.ones((5, config.hidden_size), dtype=dtype)
        y = mlp(x)
        assert y.shape == (5, config.hidden_size)
        assert y.dtype == dtype


class TestQwen2_5_VisionAttention:

    @patch('tpu_inference.models.jax.qwen2_5_vl.sharded_flash_attention')
    def test_forward_fullattn(self, mock_flash_attention: MagicMock,
                              mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                              mesh: Mesh, rng: PRNGKey):
        with jax.set_mesh(mesh):
            attn_module = Qwen2_5_VisionAttention(
                mock_vllm_config.model_config.hf_config,
                mock_vllm_config.model_config.dtype, rngs, mesh)
        B, T, D = 1, 10, attn_module.hidden_size
        # sharded_flash_attention is a factory, so we mock the returned function
        mock_attn_fn = MagicMock(return_value=jnp.ones((B,
                                                        attn_module.num_heads,
                                                        128,
                                                        attn_module.head_dim)))
        attn_module.flash_attention = mock_attn_fn
        x = jax.random.normal(rng, (T, B, D))
        rotary_pos_emb = jax.random.normal(rng, (T, attn_module.head_dim // 2))
        cu_seqlens = jnp.array([0, 5])

        y_full = attn_module(x,
                             rotary_pos_emb,
                             cu_window_seqlens=cu_seqlens,
                             use_fullattn=True)
        assert y_full.shape == (T, B, D)
        mock_attn_fn.assert_called_once()
        assert mock_attn_fn.call_args[0][3].q.shape == (1, 128)

    @patch('tpu_inference.models.jax.qwen2_5_vl.sharded_flash_attention')
    def test_forward_windowed(self, mock_flash_attention: MagicMock,
                              mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                              mesh: Mesh, rng: PRNGKey):
        with jax.set_mesh(mesh):
            attn_module = Qwen2_5_VisionAttention(
                mock_vllm_config.model_config.hf_config,
                mock_vllm_config.model_config.dtype, rngs, mesh)
        B, T, D = 1, 10, attn_module.hidden_size
        mock_attn_fn = MagicMock(return_value=jnp.ones((B,
                                                        attn_module.num_heads,
                                                        128,
                                                        attn_module.head_dim)))
        attn_module.flash_attention = mock_attn_fn
        x = jax.random.normal(rng, (T, B, D))
        rotary_pos_emb = jax.random.normal(rng, (T, attn_module.head_dim // 2))
        cu_window_seqlens = jnp.array([0, 5, 10])

        y_window = attn_module(x,
                               rotary_pos_emb,
                               cu_window_seqlens=cu_window_seqlens,
                               use_fullattn=False)
        assert y_window.shape == (T, B, D)
        mock_attn_fn.assert_called_once()
        assert mock_attn_fn.call_args[0][3].q.shape == (1, 128)

    def test_batch_fail(self, mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                        mesh: Mesh, rng: PRNGKey):
        with jax.set_mesh(mesh):
            attn_module = Qwen2_5_VisionAttention(
                mock_vllm_config.model_config.hf_config,
                mock_vllm_config.model_config.dtype, rngs, mesh)
        T, B, D = 10, 2, attn_module.hidden_size
        x = jax.random.normal(rng, (T, B, D))
        rotary_pos_emb = jax.random.normal(rng, (T, attn_module.head_dim // 2))
        with pytest.raises(
                AssertionError,
                match="Vision attention currently only supports batch size 1"):
            attn_module(x, rotary_pos_emb, use_fullattn=True)


class TestQwen2_5_VisionBlock:

    @patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2_5_VisionMLP',
           autospec=True)
    @patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2_5_VisionAttention',
           autospec=True)
    def test_forward(self, MockAttention: MagicMock, MockMLP: MagicMock,
                     mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                     mesh: Mesh, rng: PRNGKey):
        config = mock_vllm_config.model_config.hf_config
        dtype = mock_vllm_config.model_config.dtype
        D = config.vision_config.hidden_size
        T, B = 10, 1

        mock_attn_instance = MockAttention.return_value
        mock_attn_instance.return_value = jnp.zeros((T, B, D), dtype=dtype)
        mock_mlp_instance = MockMLP.return_value
        mock_mlp_instance.return_value = jnp.zeros((T, B, D), dtype=dtype)

        with jax.set_mesh(mesh):
            block = Qwen2_5_VisionBlock(config, dtype, rngs, mesh)
        x = jax.random.normal(rng, (T, B, D))
        rotary_pos_emb = jax.random.normal(
            rng, (T, config.vision_config.hidden_size //
                  config.vision_config.num_heads // 2))

        y = block(x, rotary_pos_emb, use_fullattn=True)
        assert y.shape == (T, B, D)
        mock_attn_instance.assert_called_once()
        mock_mlp_instance.assert_called_once()


class TestQwen2_5_VisionPatchEmbed:

    def test_forward(self, mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                     rng: PRNGKey, mesh: Mesh):
        vc = mock_vllm_config.model_config.hf_config.vision_config
        dtype = mock_vllm_config.model_config.dtype
        with jax.set_mesh(mesh):
            patch_embed = Qwen2_5_VisionPatchEmbed(
                rngs,
                patch_size=vc.patch_size,
                temporal_patch_size=vc.temporal_patch_size,
                in_channels=vc.in_channels,
                hidden_size=vc.hidden_size,
                dtype=dtype)
        num_patches = 4
        patch_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        x = jax.random.normal(rng, (num_patches, patch_dim))
        y = patch_embed(x)
        assert y.shape == (num_patches, vc.hidden_size)


class TestQwen2_5_VisionPatchMerger:

    def test_forward(self, mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                     rng: PRNGKey, mesh: Mesh):
        vc = mock_vllm_config.model_config.hf_config.vision_config
        dtype = mock_vllm_config.model_config.dtype
        with jax.set_mesh(mesh):
            merger = Qwen2_5_VisionPatchMerger(
                d_model=vc.out_hidden_size,
                context_dim=vc.hidden_size,
                norm_layer=partial(nnx.RMSNorm, epsilon=1e-6),
                spatial_merge_size=vc.spatial_merge_size,
                dtype=dtype,
                rngs=rngs)
        x = jax.random.normal(rng,
                              (5, vc.spatial_merge_size**2, vc.hidden_size))
        y = merger(x)
        assert y.shape == (5, vc.out_hidden_size)


class TestQwen2_5_VisionRotaryEmbedding:

    def test_forward(self, mesh: Mesh):
        dim = 16
        seqlen = 10
        with jax.set_mesh(mesh):
            rotary_emb = Qwen2_5_VisionRotaryEmbedding(dim=dim)
        emb = rotary_emb(seqlen)
        assert emb.shape == (seqlen, dim // 2)


class TestQwen2_5_VisionTransformer:

    @pytest.fixture
    def vision_transformer(self, mock_vllm_config: MockVllmConfig,
                           rngs: nnx.Rngs, mesh: Mesh):
        with jax.set_mesh(mesh):
            return Qwen2_5_VisionTransformer(mock_vllm_config, rngs, mesh)

    def test_rotary_pos_emb_thw(self,
                                vision_transformer: Qwen2_5_VisionTransformer):
        t, h, w = 2, 4, 4
        emb = vision_transformer.rotary_pos_emb_thw(t, h, w)
        vc = vision_transformer.config
        sm = vc.spatial_merge_size
        head_dim_half = (vc.hidden_size // vc.num_heads) // 2
        expected_shape = (t * (h // sm) * (w // sm), sm * sm, head_dim_half)
        assert emb.shape == expected_shape

    def test_get_window_index_thw(
            self, vision_transformer: Qwen2_5_VisionTransformer):
        grid_t, grid_h, grid_w = 1, 8, 8
        index_new, cu_seqlens_tmp = vision_transformer.get_window_index_thw(
            grid_t, grid_h, grid_w)
        vc = vision_transformer.config
        sm = vc.spatial_merge_size
        num_valid_indices = grid_t * (grid_h // sm) * (grid_w // sm)
        assert index_new.shape == (num_valid_indices, )
        assert jnp.all(index_new >= 0)

    def test_get_rope_by_thw(self,
                             vision_transformer: Qwen2_5_VisionTransformer):
        t, h, w = 1, 8, 8
        res = vision_transformer.get_rope_by_thw(t, h, w)
        assert isinstance(res, tuple)
        assert len(res) == 4
        rotary_pos_emb_thw, window_index_thw, cu_seqlens_window_thw, cu_seqlens_thw = res

        vc = vision_transformer.config
        sm = vc.spatial_merge_size
        # The rotary embedding output for each position is head_dim // 2
        head_dim_rope = (vc.hidden_size // vc.num_heads) // 2
        expected_len = window_index_thw.shape[0] * sm * sm
        assert rotary_pos_emb_thw.shape == (expected_len, head_dim_rope)

    @pytest.mark.parametrize("enable_dynamic_image_sizes", [False, True])
    def test_call(self, mock_vllm_config: MockVllmConfig, rngs: nnx.Rngs,
                  mesh: Mesh, rng: PRNGKey, enable_dynamic_image_sizes: bool):
        mock_vllm_config.additional_config = {
            "enable_dynamic_image_sizes": enable_dynamic_image_sizes
        }
        with jax.set_mesh(mesh):
            vision_transformer = Qwen2_5_VisionTransformer(
                mock_vllm_config, rngs, mesh)
        # Mock the flash_attention call to avoid sharding errors in test environment
        for block in vision_transformer.blocks:
            # The mock should return a tensor of the same shape as the query 'q'
            block.attn.flash_attention = MagicMock(
                side_effect=lambda q, k, v, seg: jnp.ones_like(q))

        vc = vision_transformer.config
        t_pix, h_pix, w_pix = 2, 84, 28

        # The number of patches is calculated from the pixel dimensions of the image/video
        num_patches = (t_pix // vc.temporal_patch_size) * \
                      (h_pix // vc.patch_size) * \
                      (w_pix // vc.patch_size)

        patch_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        x = jax.random.normal(rng, (num_patches, patch_dim))

        # The grid_thw should be in terms of patch grid dimensions, not pixels
        t_grid = t_pix // vc.temporal_patch_size
        h_grid = h_pix // vc.patch_size
        w_grid = w_pix // vc.patch_size
        grid_thw = ((t_grid, h_grid, w_grid), )

        embeddings = vision_transformer(x, grid_thw)

        # The number of output tokens is determined by the grid dimensions and spatial merge size.
        expected_len = t_grid * (h_grid // vc.spatial_merge_size) * (
            w_grid // vc.spatial_merge_size)
        assert embeddings.shape == (expected_len, vc.out_hidden_size)


class TestQwen2_5_VLForConditionalGeneration:

    @pytest.fixture
    def model(self, mock_vllm_config: MockVllmConfig, rng: PRNGKey,
              mesh: Mesh):
        with patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2_5_VisionTransformer', autospec=True) as MockVision, \
             patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2Model', autospec=True) as MockLM, jax.set_mesh(mesh):
            mock_visual = MockVision.return_value
            mock_visual.dtype = mock_vllm_config.model_config.dtype
            mock_visual.config = mock_vllm_config.model_config.hf_config.vision_config
            mock_visual.spatial_merge_size = mock_vllm_config.model_config.hf_config.vision_config.spatial_merge_size

            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
            # Directly assign mocked instances
            model.visual = mock_visual
            model.model = MockLM.return_value
            model.compute_logits = MagicMock()
            yield model

    def test_validate_and_reshape_mm_tensor(
            self, model: Qwen2_5_VLForConditionalGeneration):
        data_list = [np.ones((2, 4)), np.ones((3, 4))]
        reshaped_list = model._validate_and_reshape_mm_tensor(
            data_list, "test_list")
        assert reshaped_list.shape == (5, 4)
        assert isinstance(reshaped_list, jax.Array)

        data_2d = np.ones((5, 4))
        reshaped_2d = model._validate_and_reshape_mm_tensor(data_2d, "test_2d")
        assert reshaped_2d.shape == (5, 4)

        data_3d = np.ones((2, 5, 4))
        reshaped_3d = model._validate_and_reshape_mm_tensor(data_3d, "test_3d")
        assert reshaped_3d.shape == (10, 4)

        with pytest.raises(ValueError, match="Incorrect type of test_invalid"):
            model._validate_and_reshape_mm_tensor("invalid", "test_invalid")

    def test_parse_and_validate_image_input(
            self, model: Qwen2_5_VLForConditionalGeneration):
        grid = ((2, 28, 28), )
        vc = model.config.vision_config
        patch_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        pixel_values = np.ones((4, patch_dim))

        parsed = model._parse_and_validate_image_input(
            grid, pixel_values=pixel_values)
        assert parsed is not None
        assert parsed['type'] == "pixel_values"
        assert parsed['pixel_values'].shape == (4, patch_dim)
        assert parsed['image_grid_thw'] == grid

        parsed_none = model._parse_and_validate_image_input(grid)
        assert parsed_none is None

    def test_parse_and_validate_multimodal_inputs(
            self, model: Qwen2_5_VLForConditionalGeneration):
        grid = ((2, 28, 28), )
        vc = model.config.vision_config
        patch_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        pixel_values = np.ones((4, patch_dim))

        mm_inputs = model._parse_and_validate_multimodal_inputs(
            grid, pixel_values=pixel_values)
        assert "image" in mm_inputs
        assert mm_inputs["image"]['type'] == "pixel_values"

        mm_inputs_empty = model._parse_and_validate_multimodal_inputs(grid)
        assert not mm_inputs_empty

    def test_process_image_input_pixels(
            self, model: Qwen2_5_VLForConditionalGeneration, mesh: Mesh):
        grid_thw = ((2, 28, 28), (2, 28, 28))
        vc = model.config.vision_config
        num_patches = 8  # 4 per image
        patch_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        pixel_values = jnp.ones((num_patches, patch_dim))
        with jax.set_mesh(mesh):
            image_input = Qwen2_5_VLImagePixelInputs(type="pixel_values",
                                                     pixel_values=pixel_values,
                                                     image_grid_thw=grid_thw)

        tokens_per_image = (2 * 28 * 28) // (vc.spatial_merge_size**2)
        mock_embeds = jnp.ones((tokens_per_image, vc.out_hidden_size))
        model.visual.return_value = mock_embeds

        embeddings = model._process_image_input(image_input)
        assert isinstance(embeddings, tuple)
        assert len(embeddings) == 2
        assert embeddings[0].shape == (tokens_per_image, vc.out_hidden_size)
        assert embeddings[1].shape == (tokens_per_image, vc.out_hidden_size)
        assert model.visual.call_count == 2

    def test_embed_multimodal(self, model: Qwen2_5_VLForConditionalGeneration):
        grid_thw = ((2, 28, 28), )
        vc = model.config.vision_config
        patch_dim = vc.in_channels * vc.temporal_patch_size * vc.patch_size * vc.patch_size
        pixel_values = np.ones((4, patch_dim))
        tokens_per_image = (2 * 28 * 28) // (vc.spatial_merge_size**2)
        mock_vision_output = jnp.ones((tokens_per_image, vc.out_hidden_size))

        with patch.object(model,
                          '_process_image_input',
                          return_value=(mock_vision_output, )) as mock_process:
            mm_embeds = model.embed_multimodal(grid_thw,
                                               pixel_values=pixel_values)
            mock_process.assert_called_once()
            assert isinstance(mm_embeds, tuple)
            assert len(mm_embeds) == 1
            assert mm_embeds[0].shape == (tokens_per_image, vc.out_hidden_size)

        mm_embeds_none = model.embed_multimodal(grid_thw)
        assert len(mm_embeds_none) == 0

    @patch('tpu_inference.models.jax.qwen2_5_vl.merge_multimodal_embeddings')
    def test_embed_input_ids(self, mock_merge_embeddings: MagicMock,
                             model: Qwen2_5_VLForConditionalGeneration,
                             rng: PRNGKey):
        input_ids = jax.random.randint(rng, (1, 10), 0,
                                       model.config.vocab_size)
        mock_text_embeds = jnp.ones((1, 10, model.config.hidden_size))
        model.model = MagicMock()
        model.model.embed_tokens = MagicMock(return_value=mock_text_embeds)

        embeds = model.embed_input_ids(input_ids, None)
        np.testing.assert_array_equal(embeds, mock_text_embeds)
        mock_merge_embeddings.assert_not_called()

        empty_mm = jnp.ones((0, model.config.hidden_size), )
        embeds_empty_mm = model.embed_input_ids(input_ids, empty_mm)
        np.testing.assert_array_equal(embeds_empty_mm, mock_text_embeds)
        mock_merge_embeddings.assert_not_called()

        mm_embeds = jnp.ones((5, model.config.hidden_size))
        mock_merged = jnp.ones((1, 15, model.config.hidden_size))
        mock_merge_embeddings.return_value = mock_merged

        embeds_mm = model.embed_input_ids(input_ids, mm_embeds)
        np.testing.assert_array_equal(embeds_mm, mock_merged)
        mock_merge_embeddings.assert_called_once_with(
            input_ids, mock_text_embeds, mm_embeds,
            [model.config.image_token_id, model.config.video_token_id])

    def test_call(self, model: Qwen2_5_VLForConditionalGeneration,
                  rng: PRNGKey):
        kv_caches = [MagicMock()]
        input_ids = jax.random.randint(rng, (1, 10), 0,
                                       model.config.vocab_size)
        attn_meta = MagicMock(spec=AttentionMetadata)
        mock_lm_output = ([MagicMock()],
                          jnp.ones((1, 10, model.config.hidden_size)))
        model.model.return_value = mock_lm_output

        new_kvs, x, aux_hidden_states = model(kv_caches, input_ids, attn_meta)
        model.model.assert_called_once_with(kv_caches=kv_caches,
                                            input_ids=input_ids,
                                            attention_metadata=attn_meta,
                                            inputs_embeds=None)
        assert len(new_kvs) == 1
        assert x.shape == (1, 10, model.config.hidden_size)
        assert len(aux_hidden_states) == 0

    def test_compute_logits(self, model: Qwen2_5_VLForConditionalGeneration,
                            rng: PRNGKey):
        hidden_states = jnp.ones((1, 10, model.config.hidden_size))
        mock_logits = jnp.ones((1, 10, model.config.vocab_size))
        model.compute_logits.return_value = mock_logits

        logits = model.compute_logits(hidden_states)
        np.testing.assert_array_equal(logits, mock_logits)
        model.compute_logits.assert_called_once_with(hidden_states)

    @patch("tpu_inference.models.jax.utils.weight_utils.load_hf_weights")
    def test_load_weights(self, mock_load_weights: MagicMock,
                          model: Qwen2_5_VLForConditionalGeneration,
                          mock_vllm_config: MockVllmConfig, rng: PRNGKey,
                          mesh: Mesh):
        model.load_weights(rng)
        mock_load_weights.assert_called_once()
        kwargs = mock_load_weights.call_args.kwargs
        assert kwargs['vllm_config'] == mock_vllm_config
        assert kwargs['model'] is model
        assert kwargs['mesh'] is mesh
        assert isinstance(model.rng, nnx.Rngs)
        assert model.rng is model.rng

    @patch("tpu_inference.models.jax.utils.weight_utils.load_hf_weights")
    def test_load_weights_tied(self, mock_load_weights: MagicMock,
                               rng: PRNGKey, mesh: Mesh):
        mock_vllm_config_tied = MockVllmConfig(tie_word_embeddings=True)
        with patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2_5_VisionTransformer', autospec=True), \
             patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2Model', autospec=True), jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config_tied,
                                                       rng, mesh)
            model.load_weights(rng)
            mock_load_weights.assert_called_once()
            kwargs = mock_load_weights.call_args.kwargs
            assert "lm_head" not in kwargs['metadata_map'].name_map


class TestQwen2_5_VLPipelineParallel:

    @pytest.fixture
    def mock_pp_group(self):
        with patch('tpu_inference.models.jax.qwen2_5_vl.get_pp_group') as mock:
            yield mock

    def test_init_first_rank(self, mock_vllm_config, rng, mesh, mock_pp_group):
        mock_pp_group.return_value.is_first_rank = True
        mock_pp_group.return_value.is_last_rank = False
        with jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
            assert not isinstance(model.visual, PPMissingLayer)
            assert isinstance(model.lm_head, PPMissingLayer)
            assert model.is_first_rank is True
            assert model.is_last_rank is False

    def test_init_last_rank(self, mock_vllm_config, rng, mesh, mock_pp_group):
        mock_pp_group.return_value.is_first_rank = False
        mock_pp_group.return_value.is_last_rank = True
        with jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
            assert isinstance(model.visual, PPMissingLayer)
            assert not isinstance(model.lm_head, PPMissingLayer)
            assert model.is_first_rank is False
            assert model.is_last_rank is True

    def test_init_middle_rank(self, mock_vllm_config, rng, mesh,
                              mock_pp_group):
        mock_pp_group.return_value.is_first_rank = False
        mock_pp_group.return_value.is_last_rank = False
        with jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
            assert isinstance(model.visual, PPMissingLayer)
            assert isinstance(model.lm_head, PPMissingLayer)
            assert model.is_first_rank is False
            assert model.is_last_rank is False

    def test_call_first_rank(self, mock_vllm_config, rng, mesh, mock_pp_group):
        mock_pp_group.return_value.is_first_rank = True
        mock_pp_group.return_value.is_last_rank = False

        with patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2_5_VisionTransformer', autospec=True), \
             patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2Model', autospec=True), jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
            kv_caches = [jnp.array([])]
            input_ids = jnp.array([1, 2, 3])
            attn_meta = MagicMock(spec=AttentionMetadata)

            model.model.return_value = (kv_caches, jnp.ones((3, 16)))
            new_kvs, x, aux = model(kv_caches,
                                    input_ids,
                                    attn_meta,
                                    is_first_rank=True,
                                    is_last_rank=False)
            assert isinstance(x, JaxIntermediateTensors)
            assert "hidden_states" in x.tensors
            model.model.assert_called_once()
            _, kwargs = model.model.call_args
            np.testing.assert_array_equal(kwargs['input_ids'], input_ids)

    def test_call_non_first_rank(self, mock_vllm_config, rng, mesh,
                                 mock_pp_group):
        mock_pp_group.return_value.is_first_rank = False
        mock_pp_group.return_value.is_last_rank = True

        with patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2_5_VisionTransformer', autospec=True), \
             patch('tpu_inference.models.jax.qwen2_5_vl.Qwen2Model', autospec=True), jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
            kv_caches = [jnp.array([])]
            attn_meta = MagicMock(spec=AttentionMetadata)
            intermediate = JaxIntermediateTensors(
                tensors={"hidden_states": jnp.ones((3, 16))})

            model.model.return_value = (kv_caches, jnp.ones((3, 16)))
            new_kvs, x, aux = model(kv_caches,
                                    None,
                                    attn_meta,
                                    intermediate_tensors=intermediate,
                                    is_first_rank=False,
                                    is_last_rank=True)
            assert isinstance(x, jax.Array)
            model.model.assert_called_once()
            _, kwargs = model.model.call_args
            assert kwargs['inputs_embeds'] is intermediate["hidden_states"]

    def test_load_weights_pp_missing_layers(self, mock_vllm_config, rng, mesh,
                                            mock_pp_group):
        mock_pp_group.return_value.is_first_rank = False
        mock_pp_group.return_value.is_last_rank = False
        with jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)

        with patch(
                "tpu_inference.models.jax.utils.weight_utils.load_hf_weights"
        ) as mock_load:
            model.load_weights(rng)
            mock_load.assert_called_once()
            pp_missing = mock_load.call_args.kwargs['pp_missing_layers']
            assert "visual" in pp_missing
            assert "lm_head" in pp_missing

    def test_precompile_vision_encoder_non_first_rank(self, mock_vllm_config,
                                                      rng, mesh,
                                                      mock_pp_group):
        mock_pp_group.return_value.is_first_rank = False
        with jax.set_mesh(mesh):
            model = Qwen2_5_VLForConditionalGeneration(mock_vllm_config, rng,
                                                       mesh)
        run_compilation_fn = MagicMock()
        model.precompile_vision_encoder(run_compilation_fn)
        run_compilation_fn.assert_not_called()



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os
import time

import pytest
from vllm import LLM, SamplingParams


@pytest.fixture
def model_name():
    """Choose LLama3 8b as the test model as it supports PP on jax model impl."""
    return "meta-llama/Llama-3.1-8B-Instruct"


@pytest.fixture
def test_prompts():
    """Simple test prompts for data parallelism testing."""
    return [
        "Hello, my name is",
        "The capital of France is",
        "The colors of the rainbow are",
        "The future of AI is",
        "The president of the United States is",
        "How many players are on a standard soccer team?",
        "In Greek mythology, who is the god of the sea?",
        "What is the capital of Australia?",
        "What is the largest planet in our solar system?",
        "Who developed the theory of general relativity?",
    ]


@pytest.fixture
def sampling_params():
    """Standard sampling parameters for testing."""
    return SamplingParams(
        temperature=0.0,
        max_tokens=32,
        ignore_eos=True,
        logprobs=1,
    )


def _run_inference_with_config(model_name: str,
                               test_prompts: list,
                               sampling_params: SamplingParams,
                               tensor_parallel_size: int = 1,
                               pipeline_parallel_size: int = 1,
                               data_parallel_size: int = 1,
                               enable_expert_parallel: bool = False,
                               additional_config: dict = {},
                               kv_cache_dtype: str = "auto",
                               enable_prefix_caching: bool = False) -> list:
    """Helper function to run inference with specified configuration."""

    llm = LLM(
        model=model_name,
        max_model_len=128,
        tensor_parallel_size=tensor_parallel_size,
        pipeline_parallel_size=pipeline_parallel_size,
        data_parallel_size=data_parallel_size,
        gpu_memory_utilization=0.40,
        max_num_batched_tokens=128,
        max_num_seqs=16,
        enable_prefix_caching=enable_prefix_caching,
        additional_config=additional_config,
        kv_cache_dtype=kv_cache_dtype,
        async_scheduling=False,
        enable_expert_parallel=enable_expert_parallel,
    )

    try:
        outputs = llm.generate(test_prompts, sampling_params)
        return outputs
    finally:
        del llm
        # Wait for TPUs to be released
        time.sleep(5)


@pytest.mark.parametrize(
    "case",
    [
        {
            "id": "jax_model",
            "model": None,
            "tp": 1,
            "pp": 2,
            "dp": 1,
            "ep": False,
            "impl": "jax",
            "additional_config": {}
        },
        {
            "id": "vllm_model",
            "model": None,
            "tp": 1,
            "pp": 2,
            "dp": 1,
            "ep": False,
            "impl": "vllm",
            "additional_config": {}
        },
        {
            "id": "ep",
            "model": "Qwen/Qwen1.5-MoE-A2.7B",
            "tp": 1,
            "pp": 2,
            "dp": 1,
            "ep": True,
            "impl": "vllm",
            "additional_config": {
                "sharding": {
                    "sharding_strategy": {
                        "expert_parallelism": 2
                    }
                }
            }
        },
        {
            "id": "dp",
            "model": None,
            "tp": 1,
            "pp": 2,
            "dp": 2,
            "ep": False,
            "impl": "jax",
            "additional_config": {}
        },
        {
            "id": "tp",
            "model": None,
            "tp": 2,
            "pp": 2,
            "dp": 1,
            "ep": False,
            "impl": "jax",
            "additional_config": {}
        },
        {
            "id": "ep_tp",
            "model": "Qwen/Qwen1.5-MoE-A2.7B",
            "tp": 2,
            "pp": 2,
            "dp": 1,
            "ep": True,
            "impl": "vllm",
            "additional_config": {
                "sharding": {
                    "sharding_strategy": {
                        "expert_parallelism": 2,
                    }
                }
            }
        },
    ],
    ids=lambda x: x["id"],
)
def test_pipeline_parallel_configs(
    case,
    model_name: str,
    test_prompts: list,
    sampling_params: SamplingParams,
):
    """
    Test various pipeline parallel configurations.
    """
    impl = case["impl"]
    if impl == 'vllm':
        os.environ['MODEL_IMPL_TYPE'] = 'vllm'
    else:
        os.environ.pop('MODEL_IMPL_TYPE', None)

    model = case["model"]
    actual_model = model if model is not None else model_name

    outputs = _run_inference_with_config(
        model_name=actual_model,
        test_prompts=test_prompts,
        sampling_params=sampling_params,
        tensor_parallel_size=case["tp"],
        pipeline_parallel_size=case["pp"],
        data_parallel_size=case["dp"],
        enable_expert_parallel=case["ep"],
        additional_config=case["additional_config"],
    )

    assert len(outputs) == len(test_prompts)
    for output in outputs:
        assert len(output.outputs) > 0
        assert len(output.outputs[0].text.strip()) > 0


def _check_correctness(test_name: str, baseline_outputs: list,
                       dp_outputs: list):
    """Verify outputs match between baseline and data parallel runs."""
    assert len(baseline_outputs) == len(dp_outputs)

    text_matches = 0
    logprob_matches = 0
    total_compared_logprobs = 0
    max_logprob_diff = 0.0

    for i, (baseline, dp_result) in enumerate(zip(baseline_outputs,
                                                  dp_outputs)):
        baseline_text = baseline.outputs[0].text.strip()
        dp_text = dp_result.outputs[0].text.strip()

        # Calculate word overlap for fuzzy matching
        baseline_words = set(baseline_text.split())
        dp_words = set(dp_text.split())
        overlap = baseline_words & dp_words
        match_percent = len(overlap) / len(
            baseline_words) if baseline_words else 0

        if match_percent >= 0.7:
            text_matches += 1

        if baseline_text != dp_text:
            print(f"Text mismatch found in prompt {i}:")
            print(f"  Baseline: {baseline_text}")
            print(f"  Data Parallel: {dp_text}")
            print(f"  Match percent: {match_percent:.2%}")

        # Compare log probabilities
        baseline_logprobs = baseline.outputs[0].logprobs
        dp_logprobs = dp_result.outputs[0].logprobs

        if baseline_logprobs is None or dp_logprobs is None:
            continue

        assert len(baseline_logprobs) == len(dp_logprobs), (
            f"Logprobs length mismatch: {len(baseline_logprobs)} vs {len(dp_logprobs)}"
        )

        for token_idx, (base_lp,
                        dp_lp) in enumerate(zip(baseline_logprobs,
                                                dp_logprobs)):
            if not (base_lp and dp_lp):
                continue

            base_top_token = list(base_lp.keys())[0]
            dp_top_token = list(dp_lp.keys())[0]

            # Only compare logprobs if tokens match
            if base_top_token != dp_top_token:
                continue

            base_logprob_val = base_lp[base_top_token].logprob
            dp_logprob_val = dp_lp[dp_top_token].logprob
            diff = abs(base_logprob_val - dp_logprob_val)
            max_logprob_diff = max(max_logprob_diff, diff)
            total_compared_logprobs += 1

            if diff < 0.1:
                logprob_matches += 1
            else:
                print(f"  Logprob mismatch in prompt {i}, token {token_idx}: "
                      f"Baseline={base_logprob_val}, DP={dp_logprob_val}, "
                      f"Diff={diff:.6e}")

    # Report results
    logprob_match_rate = (logprob_matches / total_compared_logprobs
                          if total_compared_logprobs > 0 else 0)
    print(f"✓ {test_name} correctness test results:")
    print(f"  Text: {text_matches}/{len(baseline_outputs)} matches")
    print("  Target text match rate: >=60%")
    print(
        f"  Logprobs: {logprob_matches}/{total_compared_logprobs} ({logprob_match_rate:.2%}) matches (diff < 0.1)"
    )
    print(f"  Max logprob difference: {max_logprob_diff:.6e}")

    # Validate thresholds
    text_match_rate = text_matches / len(baseline_outputs)
    assert text_match_rate >= 0.6, f"Text match rate {text_match_rate:.2%} is too low"

    if total_compared_logprobs > 0:
        assert logprob_match_rate >= 0.9, f"Logprob match rate {logprob_match_rate:.2%} is too low"


def test_pipeline_parallelism_jax_model_correctness(
    model_name: str,
    test_prompts: list,
    sampling_params: SamplingParams,
):
    """
    Test that pipeline parallelism produces consistent results compared to a baseline.
    This test compares outputs from a single-device run with pipeline parallel runs
    to ensure correctness, including log probabilities.
    """
    os.environ['SKIP_JAX_PRECOMPILE'] = '1'
    os.environ['VLLM_XLA_CHECK_RECOMPILATION'] = '0'

    # Use a smaller subset of prompts for correctness testing
    small_prompts = test_prompts[:10]

    # Run baseline (no PP)
    baseline_outputs = _run_inference_with_config(
        model_name=model_name,
        test_prompts=small_prompts,
        sampling_params=sampling_params,
        tensor_parallel_size=1,
        pipeline_parallel_size=1,
    )

    # Run with model data parallelism and async scheduling
    pp_outputs = _run_inference_with_config(
        model_name=model_name,
        test_prompts=small_prompts,
        sampling_params=sampling_params,
        tensor_parallel_size=1,
        pipeline_parallel_size=2,
    )

    _check_correctness("Pipeline Parallelism", baseline_outputs, pp_outputs)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Performance tests for async copy on TPU."""

import gc
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized
from jax._src import compilation_cache as cc
from jax._src import test_util as jtu

from tpu_inference.distributed.kv_transfer import multi_layer_copy

P = jax.sharding.PartitionSpec


def create_mesh(axis_shapes, axis_names, explicit_axis: bool = False):
    """Creates a JAX device mesh with the default device order."""
    try:
        num_required_devices = np.prod(axis_shapes)
        devices = np.array(jax.devices())
        if len(devices) < num_required_devices:
            print(
                f'Expected at least {num_required_devices} devices, but only found'
                f' {len(devices)}. This script requires more devices.')
            return None

        device_array = devices[:num_required_devices].reshape(axis_shapes)
        axis_types = (tuple([jax.sharding.AxisType.Explicit] *
                            len(axis_shapes)) if explicit_axis else None)
        return jax.sharding.Mesh(device_array,
                                 axis_names,
                                 axis_types=axis_types)
    except RuntimeError:
        print('No TPU devices found. This script must be run on a TPU node.')
        return None


def create_single_layer_kv_cache(
    cache_shape: Tuple,
    cache_dtype: jnp.dtype,
    cache_sharding: jax.sharding.NamedSharding,
    init_zeros: bool = False,
) -> jax.Array:
    """Creates a single layer KV cache.

  Args:
    cache_shape: The shape of the cache.
    cache_dtype: The dtype of the cache.
    cache_sharding: The sharding specification for the cache.
    init_zeros: If True, initialize the cache with zeros. Otherwise, use random
      uniform values.

  Returns:
    A sharded jax.Array representing the KV cache.
  """

    def _allocate() -> jax.Array:
        if init_zeros:
            return jnp.zeros(shape=cache_shape, dtype=cache_dtype)
        return jax.random.uniform(jax.random.key(1),
                                  shape=cache_shape,
                                  dtype=cache_dtype)

    sharded_allocate = jax.jit(_allocate, out_shardings=cache_sharding)
    return sharded_allocate()


class KVTransferTest(jtu.JaxTestCase):

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()
        cc.reset_cache()
        jax.clear_caches()

        # Force Python GC
        gc.collect()

    @parameterized.named_parameters(
        dict(
            testcase_name='single_layer_single_block',
            kv_shape=(1024, 128, 8, 2, 64),
            num_layers=1,
            src_offsets_list=[0],
            dest_offsets_list=[0],
            chunk_sizes_list=[1],
            dtype=jnp.bfloat16,
        ),
        dict(
            testcase_name='padded_offsets_array',
            kv_shape=(1024, 128, 8, 2, 64),
            num_layers=4,
            src_offsets_list=[0, 1, 10, 11, 12, 13, 14, 15, 16, 17],
            dest_offsets_list=[0, 1, 10, 11, 12, 13, 14, 15, 16, 17],
            chunk_sizes_list=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            dtype=jnp.float32,
            num_chunks=2,
        ),
        dict(
            testcase_name='padded_offsets_array_middle_blocks',
            kv_shape=(1024, 128, 8, 2, 64),
            num_layers=4,
            src_offsets_list=[0, 1, 10, 11, 12, 13, 14, 15, 16, 17],
            dest_offsets_list=[567, 32, 10, 11, 12, 13, 14, 15, 16, 17],
            chunk_sizes_list=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            dtype=jnp.bfloat16,
            num_chunks=2,
        ),
        dict(
            testcase_name='multi_layer_variable_block',
            kv_shape=(1024, 128, 8, 2, 128),
            num_layers=10,
            src_offsets_list=[0, 1, 3, 6],
            dest_offsets_list=[0, 10, 20, 30],
            chunk_sizes_list=[1, 2, 3, 4],
            dtype=jnp.bfloat16,
        ),
    )
    def test_async_copy(
        self,
        kv_shape,
        num_layers,
        src_offsets_list,
        dest_offsets_list,
        chunk_sizes_list,
        dtype,
        num_chunks=None,
    ):
        num_devices = jax.device_count()
        axis_shapes = (1, num_devices)
        axis_names = ('data', 'model')
        explicit_sharding = False
        mesh = create_mesh(axis_shapes,
                           axis_names,
                           explicit_axis=explicit_sharding)
        partition_spec = jax.sharding.PartitionSpec(None, None, 'model')
        device_sharding = jax.sharding.NamedSharding(mesh,
                                                     partition_spec,
                                                     memory_kind='device')
        replicated_sharding = jax.sharding.NamedSharding(mesh,
                                                         P(),
                                                         memory_kind='device')

        _, bs, nh, kv_div, hd = kv_shape
        src_offsets = jnp.array(src_offsets_list, dtype=jnp.int32)
        dest_offsets = jnp.array(dest_offsets_list, dtype=jnp.int32)
        chunk_sizes = jnp.array(chunk_sizes_list, dtype=jnp.int32)
        if num_chunks is None:
            num_chunks = len(chunk_sizes_list)
        num_copy_elements = sum(chunk_sizes_list)

        poison_dest_offsets = []
        if num_chunks < len(dest_offsets_list):
            poison_dest_offsets = dest_offsets_list[num_chunks:]

        src_offsets = jax.device_put(src_offsets, replicated_sharding)
        dest_offsets = jax.device_put(dest_offsets, replicated_sharding)
        chunk_sizes = jax.device_put(chunk_sizes, replicated_sharding)
        num_chunks_array = jax.device_put(
            jnp.array([num_chunks], dtype=jnp.int32), replicated_sharding)

        kv_caches = [
            create_single_layer_kv_cache(
                kv_shape,
                dtype,
                device_sharding,
                init_zeros=True,
            ) for _ in range(num_layers)
        ]
        jax.block_until_ready(kv_caches)

        num_copy_elements = int(num_copy_elements)
        src_kv_shape = (num_copy_elements, bs, nh, kv_div, hd)
        src_kv_blocks = [
            create_single_layer_kv_cache(
                src_kv_shape,
                dtype,
                device_sharding,
                init_zeros=False,
            ) for _ in range(num_layers)
        ]
        jax.block_until_ready(src_kv_blocks)

        y = multi_layer_copy(
            src_array=src_kv_blocks,
            dest_array=kv_caches,
            src_offsets=src_offsets,
            dest_offsets=dest_offsets,
            chunk_sizes=chunk_sizes,
            num_chunks=num_chunks_array,
            mesh=mesh,
            src_sharding_spec=partition_spec,
            dest_sharding_spec=partition_spec,
            replicated_sharding_spec=P(),
        )
        for _y in y:
            _y.block_until_ready()

        self.assertEqual(y[0].sharding.memory_kind, 'device')

        # Verify correctness.
        for layer_idx in range(num_layers):
            for i in range(num_chunks):
                s_off = int(src_offsets[i])
                d_off = int(dest_offsets[i])
                c_size = int(chunk_sizes[i])
                actual = np.asarray(y[layer_idx])[d_off:d_off + c_size]
                desired = np.asarray(src_kv_blocks[layer_idx])[s_off:s_off +
                                                               c_size]
                np.testing.assert_array_equal(actual, desired)
            if poison_dest_offsets:
                for p_d_off in poison_dest_offsets:
                    np.testing.assert_array_equal(
                        np.asarray(y[layer_idx])[p_d_off], 0)


if __name__ == '__main__':
    absltest.main(testLoader=jtu.JaxTestLoader())



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import MagicMock

import jax.numpy as jnp
from flax import nnx

from tpu_inference.layers.jax.layers import DenseFFW
from tpu_inference.layers.jax.moe.moe import JaxMoE
from tpu_inference.layers.jax.transformer_block import (
    SharedExpertsTransformerBlock, TransformerBlock)


class TestTransformerBlock(unittest.TestCase):
    """Unit test suite for the JAX TransformerBlock module."""

    def test_transformer_block_dense_logic(self):
        """
        Tests the forward pass logic of a dense TransformerBlock by mocking its sub-modules.
        This test verifies the sequence of operations and residual connections.
        """
        hidden_size = 1024

        mock_pre_attn_norm = MagicMock(spec=nnx.Module)
        mock_pre_mlp_norm = MagicMock(spec=nnx.Module)

        mock_attn = MagicMock(spec=nnx.Module)
        dummy_attn_output = jnp.full((64, hidden_size),
                                     2.0,
                                     dtype=jnp.bfloat16)
        dummy_kv_cache = jnp.zeros((8, 16, 16, 128), dtype=jnp.bfloat16)
        mock_attn.return_value = (dummy_kv_cache, dummy_attn_output)

        mock_mlp = MagicMock(spec=DenseFFW)
        dummy_mlp_output = jnp.full((64, hidden_size), 3.0, dtype=jnp.bfloat16)
        mock_mlp.return_value = dummy_mlp_output

        transformer_block = TransformerBlock(
            pre_attention_norm=mock_pre_attn_norm,
            pre_mlp_norm=mock_pre_mlp_norm,
            custom_module=mock_mlp,
            attn=mock_attn,
        )

        seq_len = 64
        x = jnp.ones((seq_len, hidden_size), dtype=jnp.bfloat16)
        initial_kv_cache = MagicMock()
        attention_metadata = MagicMock()

        mock_pre_attn_norm.side_effect = lambda val: val
        mock_pre_mlp_norm.side_effect = lambda val: val

        new_kv_cache, final_output = transformer_block(
            x,
            is_prefill=True,
            kv_cache=initial_kv_cache,
            attention_metadata=attention_metadata,
        )

        mock_pre_attn_norm.assert_called_once()
        self.assertTrue(
            jnp.array_equal(mock_pre_attn_norm.call_args.args[0], x))

        mock_attn.assert_called_once_with(x, True, initial_kv_cache,
                                          attention_metadata, True)

        expected_mlp_norm_input = dummy_attn_output + x

        mock_pre_mlp_norm.assert_called_once()
        self.assertTrue(
            jnp.array_equal(mock_pre_mlp_norm.call_args.args[0],
                            expected_mlp_norm_input))

        mock_mlp.assert_called_once()
        self.assertTrue(
            jnp.array_equal(mock_mlp.call_args.args[0],
                            expected_mlp_norm_input))

        expected_final_output = dummy_mlp_output + expected_mlp_norm_input
        self.assertTrue(jnp.allclose(final_output, expected_final_output))

        self.assertTrue(jnp.array_equal(new_kv_cache, dummy_kv_cache))

    def test_shared_experts_transformer_block_logic(self):
        """Tests the forward pass logic of a SharedExpertsTransformerBlock."""
        hidden_size = 1024

        mock_pre_attn_norm = MagicMock(spec=nnx.Module)
        mock_pre_mlp_norm = MagicMock(spec=nnx.Module)

        mock_attn = MagicMock(spec=nnx.Module)
        dummy_attn_output = jnp.full((64, hidden_size),
                                     2.0,
                                     dtype=jnp.bfloat16)
        dummy_kv_cache = jnp.zeros((8, 16, 16, 128), dtype=jnp.bfloat16)
        mock_attn.return_value = (dummy_kv_cache, dummy_attn_output)

        mock_moe = MagicMock(spec=JaxMoE)
        dummy_moe_output = jnp.full((64, hidden_size), 3.0, dtype=jnp.bfloat16)
        mock_moe.return_value = dummy_moe_output

        mock_shared_experts = MagicMock(spec=DenseFFW)
        dummy_shared_experts_output = jnp.full((64, hidden_size),
                                               4.0,
                                               dtype=jnp.bfloat16)
        mock_shared_experts.return_value = dummy_shared_experts_output

        transformer_block = SharedExpertsTransformerBlock(
            pre_attention_norm=mock_pre_attn_norm,
            pre_mlp_norm=mock_pre_mlp_norm,
            custom_module=mock_moe,
            attn=mock_attn,
            shared_experts=mock_shared_experts,
        )

        seq_len = 64
        x = jnp.ones((seq_len, hidden_size), dtype=jnp.bfloat16)
        initial_kv_cache = MagicMock()
        attention_metadata = MagicMock()

        mock_pre_attn_norm.side_effect = lambda val: val
        mock_pre_mlp_norm.side_effect = lambda val: val

        new_kv_cache, final_output = transformer_block(
            x,
            is_prefill=True,
            kv_cache=initial_kv_cache,
            attention_metadata=attention_metadata,
        )
        self.assertTrue(jnp.array_equal(new_kv_cache, dummy_kv_cache))
        self.assertEqual(final_output.shape, (seq_len, hidden_size))

        self.assertEqual(mock_moe.call_count, 1)
        self.assertEqual(mock_attn.call_count, 1)
        self.assertEqual(mock_shared_experts.call_count, 1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import copy
import multiprocessing
import multiprocessing.reduction
import os
import signal
from collections import defaultdict, deque
from dataclasses import dataclass
from enum import Enum
from multiprocessing import Process
from multiprocessing.connection import Connection
from time import time
from typing import Any, Dict, List, Optional, Tuple

import cloudpickle
import torch
from vllm.config import VllmConfig
from vllm.multimodal import MULTIMODAL_REGISTRY, MultiModalRegistry
from vllm.v1.core.sched.async_scheduler import AsyncScheduler
from vllm.v1.core.sched.interface import PauseState, SchedulerInterface
from vllm.v1.core.sched.output import (CachedRequestData, GrammarOutput,
                                       SchedulerOutput)
from vllm.v1.core.sched.scheduler import Scheduler
from vllm.v1.engine import EngineCoreOutputs
from vllm.v1.kv_cache_interface import KVCacheConfig
from vllm.v1.metrics.stats import PrefixCacheStats, SchedulerStats
from vllm.v1.outputs import DraftTokenIds, ModelRunnerOutput
from vllm.v1.request import Request
from vllm.v1.structured_output import StructuredOutputManager

from tpu_inference.logger import init_logger
from tpu_inference.utils import time_function

logger = init_logger(__name__)


class SchedulerCommand(Enum):
    """Enum for scheduler worker process commands."""
    ADD_REQUEST = "add_request"
    SCHEDULE = "schedule"
    FINISH_REQUESTS = "finish_requests"
    UPDATE_DRAFT_TOKEN_IDS = "update_draft_token_ids"
    UPDATE_FROM_OUTPUT = "update_from_output"
    GET_GRAMMAR_BITMASK = "get_grammar_bitmask"
    MAKE_STATS = "make_stats"
    RESET_PREFIX_CACHE = "reset_prefix_cache"
    GET_NUM_UNFINISHED_REQUESTS = "get_num_unfinished_requests"
    HAS_FINISHED_REQUESTS = "has_finished_requests"
    GET_REQUEST_COUNTS = "get_request_counts"
    GET_TOKEN_COUNT = "get_token_count"
    PROBE_COMPUTED_BLOCKS = "probe_computed_blocks"
    RESET_ENCODER_CACHE = "reset_encoder_cache"
    SET_PAUSE_STATE = "set_pause_state"
    GET_PAUSE_STATE = "get_pause_state"
    SHUTDOWN = "shutdown"


class SchedulerWorkerError(Exception):
    """Exception raised when a scheduler worker process encounters an error."""

    def __init__(self, rank: int, message: str):
        self.rank = rank
        self.message = message
        super().__init__(f"Scheduler worker {rank} error: {message}")

    def __reduce__(self):
        """Enable proper pickling/unpickling of this exception."""
        return (self.__class__, (self.rank, self.message))


# Monkey-patch multiprocessing to use cloudpickle
# Standard pickle fails to serialize the vLLM Request object.
_original_dumps = multiprocessing.reduction.ForkingPickler.dumps
_original_loads = multiprocessing.reduction.ForkingPickler.loads


def _cloudpickle_dumps(obj, protocol=None):
    """Use cloudpickle for serialization."""
    try:
        return cloudpickle.dumps(obj, protocol=protocol)
    except Exception as e:
        obj_type = type(obj).__name__
        logger.error(f"Error pickling {obj_type}: {e}")
        if isinstance(obj, tuple) and len(obj) == 2:
            cmd, data = obj
            logger.error(
                f"Failed to pickle command: {cmd}, data type: {type(data).__name__}"
            )
        raise


def _cloudpickle_loads(data):
    """Use cloudpickle for deserialization."""
    return cloudpickle.loads(data)


def _enable_cloudpickle():
    """Enable cloudpickle for multiprocessing serialization."""
    multiprocessing.reduction.ForkingPickler.dumps = staticmethod(
        _cloudpickle_dumps)
    multiprocessing.reduction.ForkingPickler.loads = staticmethod(
        _cloudpickle_loads)


def _disable_cloudpickle():
    """Restore original pickle for multiprocessing."""
    multiprocessing.reduction.ForkingPickler.dumps = _original_dumps
    multiprocessing.reduction.ForkingPickler.loads = _original_loads


def _scheduler_worker_process(
    rank: int,
    input_conn: Connection,
    output_conn: Connection,
    vllm_config: Any,
    kv_cache_config: Any,
    structured_output_manager: Any,
    block_size: int,
    mm_registry: Any,
    include_finished_set: bool,
    log_stats: bool,
    original_scheduler_cls: type,
):
    """Worker process that manages a single scheduler instance."""
    # Initialize the scheduler in this process
    scheduler = original_scheduler_cls(
        vllm_config=vllm_config,
        kv_cache_config=kv_cache_config,
        structured_output_manager=structured_output_manager,
        block_size=block_size,
        mm_registry=mm_registry,
        include_finished_set=include_finished_set,
        log_stats=log_stats,
    )

    logger.debug(f"Scheduler worker process {rank} started")

    def _send_result(result):
        """Send result back using cloudpickle serialization."""
        output_conn.send_bytes(cloudpickle.dumps(result))

    # Process commands from the input connection
    while True:
        try:
            command, data = cloudpickle.loads(input_conn.recv_bytes())

            match command:
                case SchedulerCommand.ADD_REQUEST:
                    request = data
                    scheduler.add_request(request)
                    _send_result(None)  # Signal completion

                case SchedulerCommand.SCHEDULE:
                    output = scheduler.schedule()
                    _send_result(output)

                case SchedulerCommand.FINISH_REQUESTS:
                    request_ids, finished_status = data
                    scheduler.finish_requests(request_ids, finished_status)
                    _send_result(None)  # Signal completion

                case SchedulerCommand.UPDATE_DRAFT_TOKEN_IDS:
                    draft_token_ids = data
                    scheduler.update_draft_token_ids(draft_token_ids)
                    _send_result(None)  # Signal completion

                case SchedulerCommand.UPDATE_FROM_OUTPUT:
                    scheduler_output, model_runner_output = data
                    result = scheduler.update_from_output(
                        scheduler_output, model_runner_output)
                    _send_result(result)

                case SchedulerCommand.GET_GRAMMAR_BITMASK:
                    scheduler_output = data
                    result = scheduler.get_grammar_bitmask(scheduler_output)
                    _send_result(result)

                case SchedulerCommand.MAKE_STATS:
                    spec_decoding_stats, kv_connector_stats = data
                    result = scheduler.make_stats(spec_decoding_stats,
                                                  kv_connector_stats)
                    _send_result(result)

                case SchedulerCommand.RESET_PREFIX_CACHE:
                    reset_running_requests, reset_connector = data
                    result = scheduler.reset_prefix_cache(
                        reset_running_requests=reset_running_requests,
                        reset_connector=reset_connector)
                    _send_result(result)

                case SchedulerCommand.RESET_ENCODER_CACHE:
                    scheduler.reset_encoder_cache()
                    _send_result(None)

                case SchedulerCommand.GET_NUM_UNFINISHED_REQUESTS:
                    result = scheduler.get_num_unfinished_requests()
                    _send_result(result)

                case SchedulerCommand.HAS_FINISHED_REQUESTS:
                    result = scheduler.has_finished_requests()
                    _send_result(result)

                case SchedulerCommand.GET_REQUEST_COUNTS:
                    running = len(scheduler.running)
                    waiting = len(scheduler.waiting)
                    _send_result((running, waiting))

                case SchedulerCommand.GET_TOKEN_COUNT:
                    # Calculate total tokens across running and waiting requests
                    total_tokens = 0
                    for req in scheduler.running:
                        total_tokens += len(req.all_token_ids)
                    for req in scheduler.waiting:
                        total_tokens += len(req.all_token_ids)
                    _send_result(total_tokens)

                case SchedulerCommand.PROBE_COMPUTED_BLOCKS:
                    # Probe for cached blocks without recording prefix cache stats.
                    request = data
                    kv_cache_mgr = scheduler.kv_cache_manager
                    if not kv_cache_mgr.enable_caching or request.skip_reading_prefix_cache:
                        _send_result(0)
                    else:
                        max_cache_hit_length = request.num_tokens - 1
                        _, num_cached_tokens = (
                            kv_cache_mgr.coordinator.find_longest_cache_hit(
                                request.block_hashes, max_cache_hit_length))
                        _send_result(num_cached_tokens)

                case SchedulerCommand.SET_PAUSE_STATE:
                    pause_state = data
                    scheduler.set_pause_state(pause_state)
                    _send_result(None)

                case SchedulerCommand.GET_PAUSE_STATE:
                    result = scheduler.pause_state
                    _send_result(result)

                case SchedulerCommand.SHUTDOWN:
                    logger.info(f"Rank {rank}: Shutting down")
                    scheduler.shutdown()
                    _send_result(None)  # Signal completion
                    os._exit(0)
                case _:
                    error = SchedulerWorkerError(
                        rank, f"Unknown command: {command}")
                    _send_result(error)
                    raise error

        except (SystemExit, KeyboardInterrupt):
            logger.info(f"Scheduler worker {rank} received shutdown signal, "
                        "exiting gracefully.")
            try:
                scheduler.shutdown()
            except Exception:
                pass
            os._exit(0)

        except Exception as e:
            logger.error(
                f"Error in scheduler worker {rank}: {e}. If "
                "async scheduling is enabled, consider disabling it to "
                "debug this issue.",
                exc_info=True)

            error = SchedulerWorkerError(rank, str(e))
            _send_result(error)


@dataclass
class DPSchedulerOutput(SchedulerOutput):
    """Extended SchedulerOutput that includes DP rank assignments."""
    assigned_dp_rank: Optional[Dict[str, int]] = None
    # The maximum number of tokens scheduled on any single DP rank in this step.
    # This is used by the Runner to calculate the global padded batch size
    # (padded_max * dp_size), ensuring consistent shapes across pipeline stages.
    max_num_scheduled_tokens_per_dp_rank: int = 0

    def __init__(self,
                 *args,
                 assigned_dp_rank=None,
                 max_num_scheduled_tokens_per_dp_rank=0,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.assigned_dp_rank = assigned_dp_rank or {}
        self.max_num_scheduled_tokens_per_dp_rank = max_num_scheduled_tokens_per_dp_rank


class DPScheduler(SchedulerInterface):
    """
    DPScheduler is used when DP size is >=2. Otherwise the default vLLM scheduler is used.

    The DPScheduler manages:
    1. Multiple vLLM Schedulers (one per DP rank)
    2. Request-to-scheduler assignment

    Each Scheduler manages its own logical KV cache shard and scheduling logic.

    **Load Balancing**

    For new requests:
    - If there is prefix cache hit, assigns request to the rank with the best hit
    - Otherwise, assigns request to the rank with the least total tokens

    Once a DP rank is assigned to a request, it remains fixed for the request's lifetime.
    A request will be freed from its assigned rank when it is completed or preempted.
    """

    def __init__(
        self,
        vllm_config: VllmConfig,
        kv_cache_config: KVCacheConfig,
        structured_output_manager: StructuredOutputManager,
        block_size: int,
        mm_registry: MultiModalRegistry = MULTIMODAL_REGISTRY,
        include_finished_set: bool = False,
        log_stats: bool = False,
    ) -> None:
        self.vllm_config = vllm_config
        self.block_size = block_size
        self.log_stats = log_stats
        self.connector = None
        self.structured_output_manager = structured_output_manager

        # DP state
        self.dp_size = vllm_config.sharding_config.total_dp_size
        self.assigned_dp_rank: Dict[str, int] = {}  # req_id -> dp_rank
        self.cached_schedulers_output = deque()
        self._create_per_rank_configs(kv_cache_config)

        # Initialize NONE_HASH global before forking worker processes
        # This ensures all workers inherit the initialized value
        if vllm_config.cache_config.enable_prefix_caching:
            from vllm.utils.hashing import get_hash_fn_by_name
            from vllm.v1.core.kv_cache_utils import init_none_hash
            caching_hash_fn = get_hash_fn_by_name(
                vllm_config.cache_config.prefix_caching_hash_algo)
            init_none_hash(caching_hash_fn)

        # The original scheduler class could be Scheduler or AsyncScheduler
        original_scheduler_cls = vllm_config.scheduler_config._original_scheduler_cls

        # Enable cloudpickle for multiprocessing to handle local functions
        _enable_cloudpickle()

        # Create worker processes with Pipe connections (no feeder threads).
        # multiprocessing.Queue uses background feeder threads for put(),
        # which causes GIL contention and thread convoy effects at high DP
        # sizes. Using raw Pipe connections eliminates all background threads.
        ctx = multiprocessing.get_context('fork')
        self.input_conns: List[Connection] = []  # parent writes, child reads
        self.output_conns: List[Connection] = []  # child writes, parent reads
        self.processes: List[Process] = []

        for rank in range(self.dp_size):
            # Each pipe gives (parent_end, child_end)
            # Input pipe: parent sends commands, child receives
            input_parent_conn, input_child_conn = ctx.Pipe()
            # Output pipe: child sends results, parent receives
            output_parent_conn, output_child_conn = ctx.Pipe()

            self.input_conns.append(input_parent_conn)
            self.output_conns.append(output_parent_conn)

            process = ctx.Process(
                target=_scheduler_worker_process,
                args=(
                    rank,
                    input_child_conn,
                    output_child_conn,
                    self.vllm_config,
                    self.per_rank_kv_cache_configs[rank],
                    structured_output_manager,
                    block_size,
                    mm_registry,
                    include_finished_set,
                    log_stats,
                    original_scheduler_cls,
                ),
            )
            process.start()
            # Close child ends in parent process
            input_child_conn.close()
            output_child_conn.close()
            self.processes.append(process)

        logger.info(
            f"DPScheduler (Async = {self.vllm_config.scheduler_config.async_scheduling}) "
            f"started {self.dp_size} worker processes with cloudpickle. "
            f"Per-rank limits: max_seqs={self.vllm_config.scheduler_config.max_num_seqs}, "
            f"max_tokens={self.vllm_config.scheduler_config.max_num_batched_tokens}"
        )

        # Register an atexit handler that runs *before* multiprocessing's
        # _exit_function (atexit handlers run LIFO). This kills workers
        # if shutdown() was never called (e.g. unhandled exception).
        atexit.register(self._atexit_cleanup)

    def _atexit_cleanup(self) -> None:
        """Kill worker processes if shutdown() was not called."""
        for process in self.processes:
            if process.is_alive():
                try:
                    os.kill(process.pid, signal.SIGKILL)
                except OSError:
                    pass
        for process in self.processes:
            process.join(timeout=1.0)
        multiprocessing.active_children()

    def _create_per_rank_configs(self, kv_cache_config: KVCacheConfig) -> None:
        self.per_rank_kv_cache_configs: List[KVCacheConfig] = []
        for _ in range(self.dp_size):
            rank_config = copy.deepcopy(kv_cache_config)
            rank_config.num_blocks = kv_cache_config.num_blocks // self.dp_size
            self.per_rank_kv_cache_configs.append(rank_config)

    def _send_command(self,
                      rank: int,
                      command: SchedulerCommand,
                      data: Any = None) -> None:
        """Send a command to a worker process via its input pipe."""
        start_time = time()
        payload = cloudpickle.dumps((command, data))
        serialize_time = time() - start_time
        self.input_conns[rank].send_bytes(payload)
        if serialize_time > 1.0:
            logger.warning(f"Slow serialization ({serialize_time:.2f}s, "
                           f"{len(payload)} bytes) for '{command.value}' "
                           f"command to rank {rank}/{self.dp_size}.")

    def _get_result(self,
                    rank: int,
                    command: Optional[SchedulerCommand] = None) -> Any:
        """Get result from a worker process via its output pipe.

        Uses raw Connection.recv_bytes() + cloudpickle.loads() instead of
        multiprocessing.Queue.get(). This eliminates the background feeder
        threads that Queue uses, avoiding GIL contention and thread convoy
        effects at high DP sizes.
        """
        try:
            start_time = time()
            raw_bytes = self.output_conns[rank].recv_bytes()
            recv_time = time()
            result = cloudpickle.loads(raw_bytes)
            end_time = time()
            total_time = end_time - start_time
            if total_time > 1.0:
                cmd_name = command.value if command else "unknown"
                pipe_wait = recv_time - start_time
                deserialize = end_time - recv_time
                logger.warning(f"Long wait time ({total_time:.2f}s) for "
                               f"rank {rank}/{self.dp_size} response to "
                               f"'{cmd_name}' command "
                               f"(pipe_wait={pipe_wait:.2f}s, "
                               f"deserialize={deserialize:.2f}s, "
                               f"{len(raw_bytes)} bytes).")
        except Exception as e:
            # Check if the worker process is still alive for a better message
            proc = self.processes[rank]
            if not proc.is_alive():
                exit_code = proc.exitcode
                raise RuntimeError(
                    f"Pipe error for rank {rank}: "
                    f"Worker process terminated with exit code {exit_code}. "
                    "This may indicate a crash or signal in the scheduler "
                    "worker process.") from e
            raise RuntimeError(
                f"Pipe error for rank {rank}: "
                "Worker process terminated unexpectedly. "
                "This may indicate a crash in the scheduler worker process."
            ) from e
        if isinstance(result, SchedulerWorkerError):
            raise result
        return result

    def _get_rank_token_counts(self) -> Dict[int, int]:
        """Calculate total tokens currently assigned to each DP rank."""
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.GET_TOKEN_COUNT)

        rank_tokens = {}
        for rank in range(self.dp_size):
            rank_tokens[rank] = self._get_result(
                rank, SchedulerCommand.GET_TOKEN_COUNT)

        return rank_tokens

    def _find_best_rank_for_request(self, request: Request) -> int:
        """Find the best DP rank for a new request based on load balancing."""
        rank_tokens = self._get_rank_token_counts()

        # First, try to find a rank with prefix cache hit.
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.PROBE_COMPUTED_BLOCKS,
                               request)

        best_cache_rank = None
        best_cache_tokens = 0
        for rank in range(self.dp_size):
            cached_tokens = self._get_result(
                rank, SchedulerCommand.PROBE_COMPUTED_BLOCKS)
            if cached_tokens > best_cache_tokens:
                best_cache_tokens = cached_tokens
                best_cache_rank = rank
        if best_cache_tokens > 0:
            return best_cache_rank

        # Otherwise, find rank with least tokens
        selected_rank = min(rank_tokens, key=rank_tokens.get)
        return selected_rank

    def add_request(self, request: Request) -> None:
        """
        Add a new request to the appropriate DP rank scheduler.

        This is the main entry point for new requests. The scheduler will:
        1. Determine the best DP rank for the request (load balancing + cache hits)
        2. Assign the request to that rank
        3. Add the request to the rank's scheduler
        """
        assert request.request_id not in self.assigned_dp_rank, (
            f"Request {request.request_id} already "
            f"assigned to rank {self.assigned_dp_rank[request.request_id]})")
        rank = self._find_best_rank_for_request(request)
        self.assigned_dp_rank[request.request_id] = rank

        self._send_command(rank, SchedulerCommand.ADD_REQUEST, request)
        self._get_result(rank, SchedulerCommand.ADD_REQUEST)

    @time_function
    def schedule(self) -> DPSchedulerOutput:
        """
        Main scheduling method that coordinates all DP rank schedulers.

        Process:
        1. Add any new requests to appropriate DP ranks
        2. Run each scheduler independently in parallel
        3. Combine outputs from all schedulers
        4. Return unified scheduling result
        """
        # Run each scheduler independently
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.SCHEDULE)

        # Collect outputs from all workers (blocking)
        rank_outputs = []
        for rank in range(self.dp_size):
            output = self._get_result(rank, SchedulerCommand.SCHEDULE)
            rank_outputs.append(output)

        # Cache scheduler outputs to use in `update_from_output`
        self.cached_schedulers_output.append(rank_outputs)

        # Return combined scheduler outputs
        combined_output = self._combine_scheduler_outputs(rank_outputs)

        logger.debug(
            f"DPScheduler scheduled: "
            f"{combined_output.total_num_scheduled_tokens} total tokens, "
            f"{len(combined_output.scheduled_new_reqs)} new requests, "
            f"{len(combined_output.scheduled_cached_reqs.req_ids)} cached requests"
        )

        return combined_output

    def _combine_scheduler_outputs(
            self, rank_outputs: List[SchedulerOutput]) -> DPSchedulerOutput:
        """Combine outputs from all DP rank schedulers into a unified output."""

        # Combine new requests
        all_new_reqs = []
        for output in rank_outputs:
            all_new_reqs.extend(output.scheduled_new_reqs)

        # Combine cached request data
        combined_cached_data = self._combine_cached_request_data(rank_outputs)

        # Combine token counts and other metrics
        combined_num_scheduled_tokens = {}
        combined_spec_decode_tokens = {}
        combined_encoder_inputs = {}
        total_scheduled_tokens = 0
        max_scheduled_tokens_per_rank = 0

        for output in rank_outputs:
            combined_num_scheduled_tokens.update(output.num_scheduled_tokens)
            combined_spec_decode_tokens.update(
                output.scheduled_spec_decode_tokens)
            combined_encoder_inputs.update(output.scheduled_encoder_inputs)
            total_scheduled_tokens += output.total_num_scheduled_tokens
            max_scheduled_tokens_per_rank = max(
                max_scheduled_tokens_per_rank,
                output.total_num_scheduled_tokens)

        # Combine finished request IDs
        combined_finished_req_ids = set()
        for output in rank_outputs:
            combined_finished_req_ids.update(output.finished_req_ids)

        # Combine other fields (take from first non-empty or use defaults)
        num_common_prefix_blocks = rank_outputs[
            0].num_common_prefix_blocks if rank_outputs else []

        # Create DP rank assignment mapping for scheduled requests
        assigned_dp_rank = {}
        for req_id in combined_num_scheduled_tokens.keys():
            assigned_dp_rank[req_id] = self.assigned_dp_rank[req_id]

        return DPSchedulerOutput(
            scheduled_new_reqs=all_new_reqs,
            scheduled_cached_reqs=combined_cached_data,
            num_scheduled_tokens=combined_num_scheduled_tokens,
            total_num_scheduled_tokens=total_scheduled_tokens,
            scheduled_spec_decode_tokens=combined_spec_decode_tokens,
            scheduled_encoder_inputs=combined_encoder_inputs,
            num_common_prefix_blocks=num_common_prefix_blocks,
            finished_req_ids=combined_finished_req_ids,
            free_encoder_mm_hashes=set(),
            assigned_dp_rank=assigned_dp_rank,
            max_num_scheduled_tokens_per_dp_rank=max_scheduled_tokens_per_rank,
        )

    def _combine_cached_request_data(
            self, rank_outputs: List[SchedulerOutput]) -> CachedRequestData:
        """Combine cached request data from all DP rank schedulers."""
        combined_req_ids = []
        combined_resumed_req_ids = []
        combined_new_token_ids = []
        combined_all_token_ids = {}
        combined_new_block_ids = []
        combined_num_computed_tokens = []
        combined_num_output_tokens = []

        for output in rank_outputs:
            cached_data = output.scheduled_cached_reqs

            combined_req_ids.extend(cached_data.req_ids)
            combined_resumed_req_ids.extend(cached_data.resumed_req_ids)
            combined_new_token_ids.extend(cached_data.new_token_ids)
            combined_all_token_ids.update(cached_data.all_token_ids)
            combined_new_block_ids.extend(cached_data.new_block_ids)
            combined_num_computed_tokens.extend(
                cached_data.num_computed_tokens)
            combined_num_output_tokens.extend(cached_data.num_output_tokens)

        return CachedRequestData(
            req_ids=combined_req_ids,
            resumed_req_ids=combined_resumed_req_ids,
            new_token_ids=combined_new_token_ids,
            all_token_ids=combined_all_token_ids,
            new_block_ids=combined_new_block_ids,
            num_computed_tokens=combined_num_computed_tokens,
            num_output_tokens=combined_num_output_tokens,
        )

    def _combine_scheduler_stats(
        self,
        rank_stats_list: List[Optional[SchedulerStats]],
    ) -> Optional[SchedulerStats]:
        """Combine SchedulerStats from all DP rank schedulers.

        The per-rank stats are extracted from the workers' update_from_output
        results, where the base scheduler's make_stats() already collected
        and reset the prefix cache stats.
        """
        total_running_reqs = 0
        total_waiting_reqs = 0
        total_kv_cache_usage = 0.0

        combined_prefix_cache_stats = PrefixCacheStats()
        combined_connector_prefix_cache_stats: Optional[
            PrefixCacheStats] = None
        has_any_stats = False

        for rank_stats in rank_stats_list:
            if rank_stats is None:
                continue
            has_any_stats = True

            total_running_reqs += rank_stats.num_running_reqs
            total_waiting_reqs += rank_stats.num_waiting_reqs
            total_kv_cache_usage += rank_stats.kv_cache_usage

            # Combine prefix cache stats
            if rank_stats.prefix_cache_stats:
                combined_prefix_cache_stats.reset = (
                    combined_prefix_cache_stats.reset
                    or rank_stats.prefix_cache_stats.reset)
                combined_prefix_cache_stats.requests += (
                    rank_stats.prefix_cache_stats.requests)
                combined_prefix_cache_stats.queries += (
                    rank_stats.prefix_cache_stats.queries)
                combined_prefix_cache_stats.hits += (
                    rank_stats.prefix_cache_stats.hits)

            # Combine connector prefix cache stats
            if rank_stats.connector_prefix_cache_stats:
                if combined_connector_prefix_cache_stats is None:
                    combined_connector_prefix_cache_stats = PrefixCacheStats()
                combined_connector_prefix_cache_stats.reset = (
                    rank_stats.connector_prefix_cache_stats.reset)
                combined_connector_prefix_cache_stats.requests += (
                    rank_stats.connector_prefix_cache_stats.requests)
                combined_connector_prefix_cache_stats.queries += (
                    rank_stats.connector_prefix_cache_stats.queries)
                combined_connector_prefix_cache_stats.hits += (
                    rank_stats.connector_prefix_cache_stats.hits)

        if not has_any_stats:
            return None

        # Average KV cache usage across ranks
        num_ranks = len(rank_stats_list)
        avg_kv_cache_usage = (total_kv_cache_usage /
                              num_ranks if num_ranks else 0.0)

        return SchedulerStats(
            num_running_reqs=total_running_reqs,
            num_waiting_reqs=total_waiting_reqs,
            kv_cache_usage=avg_kv_cache_usage,
            prefix_cache_stats=combined_prefix_cache_stats,
            connector_prefix_cache_stats=combined_connector_prefix_cache_stats,
        )

    def get_grammar_bitmask(
        self,
        scheduler_output: DPSchedulerOutput,
    ) -> GrammarOutput | None:
        """
        Generate grammar bitmask for structured output requests across all DP ranks.

        This method calls get_grammar_bitmask on each underlying scheduler and
        combines their outputs, similar to how other operations are handled.
        """
        # Use the most recent cached outputs from the schedule() call
        if not self.cached_schedulers_output:
            return None

        rank_scheduler_outputs = self.cached_schedulers_output[
            -1]  # Get the most recent

        combined_structured_output_request_ids = []
        combined_bitmasks = []

        # Get grammar bitmask from each DP rank scheduler
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.GET_GRAMMAR_BITMASK,
                               rank_scheduler_outputs[rank])
        for rank in range(self.dp_size):
            grammar_output = self._get_result(
                rank, SchedulerCommand.GET_GRAMMAR_BITMASK)
            if grammar_output is not None:
                combined_structured_output_request_ids.extend(
                    grammar_output.structured_output_request_ids)
                combined_bitmasks.append(grammar_output.grammar_bitmask)

        if not combined_structured_output_request_ids:
            return None

        # Combine bitmasks - concatenate along the batch dimension
        if len(combined_bitmasks) == 1:
            combined_bitmask = combined_bitmasks[0]
        else:
            combined_bitmask = torch.cat(combined_bitmasks, dim=0)

        return GrammarOutput(combined_structured_output_request_ids,
                             combined_bitmask)

    def update_from_output(
        self, scheduler_output: DPSchedulerOutput,
        model_runner_output: ModelRunnerOutput
    ) -> dict[int, EngineCoreOutputs]:
        """
        Update all DP rank schedulers based on model runner output.

        We need to route the model runner output to the appropriate scheduler
        based on which rank each request belongs to.
        """
        # Split model output by DP rank (each rank gets only its req_ids).
        rank_model_outputs = self._split_model_output_by_rank(
            model_runner_output)
        rank_scheduler_outputs = self.cached_schedulers_output.popleft()

        # Send each rank its scheduler output + per-rank model output.
        for rank in range(self.dp_size):
            self._send_command(
                rank, SchedulerCommand.UPDATE_FROM_OUTPUT,
                (rank_scheduler_outputs[rank], rank_model_outputs[rank]))

        combined_engine_outputs = defaultdict(list)
        rank_scheduler_stats: List[Optional[SchedulerStats]] = []
        for rank in range(self.dp_size):
            rank_engine_outputs = self._get_result(
                rank, SchedulerCommand.UPDATE_FROM_OUTPUT)
            rank_stats = None
            for client_idx, engine_output in rank_engine_outputs.items():
                combined_engine_outputs[client_idx].append(engine_output)
                if engine_output.scheduler_stats is not None:
                    rank_stats = engine_output.scheduler_stats
            rank_scheduler_stats.append(rank_stats)

        # Combine scheduler stats from all DP ranks
        combined_stats = self._combine_scheduler_stats(rank_scheduler_stats)

        # Clean up finished requests from DP tracking
        self._cleanup_finished_requests(scheduler_output.finished_req_ids)

        # Return combined EngineCoreOutput
        stats_attached = False
        for client_idx, engine_outputs in combined_engine_outputs.items():
            combined_output = EngineCoreOutputs()
            outputs = []
            finished_requests = set()
            for engine_output in engine_outputs:
                outputs.extend(engine_output.outputs)
                if engine_output.finished_requests:
                    finished_requests.update(engine_output.finished_requests)
            combined_output.engine_index = engine_outputs[0].engine_index
            combined_output.outputs = outputs
            combined_output.finished_requests = finished_requests
            # Attach combined stats to only the first client output
            # (matching the base scheduler behavior)
            if not stats_attached and combined_stats is not None:
                combined_output.scheduler_stats = combined_stats
                stats_attached = True
            combined_engine_outputs[client_idx] = combined_output

        return combined_engine_outputs

    def _split_model_output_by_rank(
            self,
            global_model_output: ModelRunnerOutput) -> List[ModelRunnerOutput]:
        """Split the model runner output by DP rank for individual scheduler updates."""
        outputs = [
            ModelRunnerOutput(
                req_ids=[],
                req_id_to_index=global_model_output.req_id_to_index,
                sampled_token_ids=global_model_output.sampled_token_ids,
                logprobs=global_model_output.logprobs,
                prompt_logprobs_dict=global_model_output.prompt_logprobs_dict,
                pooler_output=None,
                num_nans_in_logits=global_model_output.num_nans_in_logits,
                kv_connector_output=global_model_output.kv_connector_output,
            ) for _ in range(self.dp_size)
        ]

        for req_id in global_model_output.req_ids:
            rank = self.assigned_dp_rank[req_id]
            outputs[rank].req_ids.append(req_id)

        return outputs

    def _cleanup_finished_requests(self, finished_req_ids: set[str]) -> None:
        """Remove finished requests from our DP rank assignment tracking."""
        for req_id in finished_req_ids:
            if req_id in self.assigned_dp_rank:
                del self.assigned_dp_rank[req_id]

    def finish_requests(self, request_ids, finished_status) -> None:
        """Forward request finish signals to the appropriate DP rank schedulers."""
        if isinstance(request_ids, str):
            request_ids = [request_ids]
        elif request_ids is None:
            # None means finish all requests (matches base scheduler behavior)
            request_ids = list(self.assigned_dp_rank.keys())

        # Route finish signals to appropriate schedulers
        rank_request_ids = defaultdict(list)
        for req_id in request_ids:
            if req_id not in self.assigned_dp_rank:
                continue
            rank = self.assigned_dp_rank[req_id]
            rank_request_ids[rank].append(req_id)

        # Forward to each scheduler
        for rank, req_ids in rank_request_ids.items():
            self._send_command(rank, SchedulerCommand.FINISH_REQUESTS,
                               (req_ids, finished_status))
            self._get_result(rank, SchedulerCommand.FINISH_REQUESTS)

    def get_num_unfinished_requests(self) -> int:
        """Get total number of unfinished requests across all DP ranks."""
        for rank in range(self.dp_size):
            self._send_command(rank,
                               SchedulerCommand.GET_NUM_UNFINISHED_REQUESTS)

        total = 0
        for rank in range(self.dp_size):
            count = self._get_result(
                rank, SchedulerCommand.GET_NUM_UNFINISHED_REQUESTS)
            total += count
        return total

    def has_finished_requests(self) -> bool:
        """Check if any DP rank has finished requests."""
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.HAS_FINISHED_REQUESTS)

        has_finished_any = False
        for rank in range(self.dp_size):
            has_finished_any |= self._get_result(
                rank, SchedulerCommand.HAS_FINISHED_REQUESTS)
        return has_finished_any

    def get_request_counts(self) -> Tuple[int, int]:
        """Get total (running, waiting) request counts across all DP ranks."""
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.GET_REQUEST_COUNTS)

        total_running = 0
        total_waiting = 0
        for rank in range(self.dp_size):
            running, waiting = self._get_result(
                rank, SchedulerCommand.GET_REQUEST_COUNTS)
            total_running += running
            total_waiting += waiting
        return total_running, total_waiting

    def reset_prefix_cache(
        self,
        reset_running_requests: bool = False,
        reset_connector: bool = False,
    ) -> bool:
        """Reset prefix cache for all DP rank schedulers."""
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.RESET_PREFIX_CACHE,
                               (reset_running_requests, reset_connector))

        all_success = True
        for rank in range(self.dp_size):
            success = self._get_result(rank,
                                       SchedulerCommand.RESET_PREFIX_CACHE)
            all_success &= success
        return all_success

    def reset_encoder_cache(self) -> None:
        """Reset encoder cache for all DP rank schedulers."""
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.RESET_ENCODER_CACHE)

        for rank in range(self.dp_size):
            self._get_result(rank, SchedulerCommand.RESET_ENCODER_CACHE)

    @property
    def pause_state(self) -> PauseState:
        """Get the pause state from the first DP rank scheduler.

        All ranks share the same pause state, so we only need to query one.
        """
        self._send_command(0, SchedulerCommand.GET_PAUSE_STATE)
        return self._get_result(0, SchedulerCommand.GET_PAUSE_STATE)

    def set_pause_state(self, pause_state: PauseState) -> None:
        """Set pause state for all DP rank schedulers."""
        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.SET_PAUSE_STATE,
                               pause_state)

        for rank in range(self.dp_size):
            self._get_result(rank, SchedulerCommand.SET_PAUSE_STATE)

    def make_stats(self,
                   spec_decoding_stats=None,
                   kv_connector_stats=None) -> Optional[SchedulerStats]:
        """Combine stats from all DP rank schedulers."""
        if not self.log_stats:
            return None

        # Aggregate stats from all schedulers
        total_running_reqs = 0
        total_waiting_reqs = 0
        total_kv_cache_usage = 0.0

        combined_prefix_cache_stats = PrefixCacheStats()
        combined_connector_prefix_cache_stats: Optional[
            PrefixCacheStats] = None

        for rank in range(self.dp_size):
            self._send_command(rank, SchedulerCommand.MAKE_STATS,
                               (spec_decoding_stats, kv_connector_stats))

        for rank in range(self.dp_size):
            rank_stats = self._get_result(rank, SchedulerCommand.MAKE_STATS)
            if rank_stats is None:
                continue

            total_running_reqs += rank_stats.num_running_reqs
            total_waiting_reqs += rank_stats.num_waiting_reqs
            total_kv_cache_usage += rank_stats.kv_cache_usage

            # Combine prefix cache stats
            if rank_stats.prefix_cache_stats:
                combined_prefix_cache_stats.reset = rank_stats.prefix_cache_stats.reset
                combined_prefix_cache_stats.requests += rank_stats.prefix_cache_stats.requests
                combined_prefix_cache_stats.queries += rank_stats.prefix_cache_stats.queries
                combined_prefix_cache_stats.hits += rank_stats.prefix_cache_stats.hits

            # Combine connector prefix cache stats
            if rank_stats.connector_prefix_cache_stats:
                if combined_connector_prefix_cache_stats is None:
                    combined_connector_prefix_cache_stats = PrefixCacheStats()
                combined_connector_prefix_cache_stats.reset = rank_stats.connector_prefix_cache_stats.reset
                combined_connector_prefix_cache_stats.requests += rank_stats.connector_prefix_cache_stats.requests
                combined_connector_prefix_cache_stats.queries += rank_stats.connector_prefix_cache_stats.queries
                combined_connector_prefix_cache_stats.hits += rank_stats.connector_prefix_cache_stats.hits

        # Average KV cache usage across ranks
        avg_kv_cache_usage = total_kv_cache_usage / self.dp_size if self.dp_size else 0.0

        return SchedulerStats(
            num_running_reqs=total_running_reqs,
            num_waiting_reqs=total_waiting_reqs,
            kv_cache_usage=avg_kv_cache_usage,
            prefix_cache_stats=combined_prefix_cache_stats,
            connector_prefix_cache_stats=combined_connector_prefix_cache_stats,
            spec_decoding_stats=spec_decoding_stats,
            kv_connector_stats=kv_connector_stats.data
            if kv_connector_stats else None,
        )

    def update_draft_token_ids(self, draft_token_ids) -> None:
        """Forward draft token updates to the appropriate DP rank schedulers."""
        # Group draft tokens by DP rank based on request assignments
        rank_draft_tokens = defaultdict(lambda: {
            "req_ids": [],
            "draft_token_ids": []
        })

        for req_id, tokens in zip(draft_token_ids.req_ids,
                                  draft_token_ids.draft_token_ids):
            if req_id in self.assigned_dp_rank:
                rank = self.assigned_dp_rank[req_id]
                rank_draft_tokens[rank]["req_ids"].append(req_id)
                rank_draft_tokens[rank]["draft_token_ids"].append(tokens)

        for rank, draft_data in rank_draft_tokens.items():
            # Create a draft_token_ids object for this rank (mock structure)
            rank_draft_token_ids = type(draft_token_ids)(
                req_ids=draft_data["req_ids"],
                draft_token_ids=draft_data["draft_token_ids"])
            self._send_command(rank, SchedulerCommand.UPDATE_DRAFT_TOKEN_IDS,
                               rank_draft_token_ids)
            self._get_result(rank, SchedulerCommand.UPDATE_DRAFT_TOKEN_IDS)

    def update_draft_token_ids_in_output(
            self, draft_token_ids: "DraftTokenIds",
            scheduler_output: "SchedulerOutput") -> None:
        """Not implemented for DPScheduler."""
        raise NotImplementedError(
            "update_draft_token_ids_in_output is not implemented for DPScheduler."
        )

    def shutdown(self) -> None:
        """Shutdown all DP rank scheduler worker processes."""
        atexit.unregister(self._atexit_cleanup)

        # Send shutdown command to all workers, skipping dead ones
        for rank in range(self.dp_size):
            if not self.processes[rank].is_alive():
                logger.warning(
                    f"Rank {rank}: Worker process already terminated "
                    f"(exit code {self.processes[rank].exitcode}), "
                    "skipping shutdown command.")
                continue
            self._send_command(rank, SchedulerCommand.SHUTDOWN)

        # Wait for acknowledgment (blocking), skipping dead ones
        for rank in range(self.dp_size):
            if not self.processes[rank].is_alive():
                continue
            try:
                self._get_result(rank, SchedulerCommand.SHUTDOWN)
            except (RuntimeError, OSError) as e:
                logger.warning(
                    f"Rank {rank}: Failed to get shutdown acknowledgment: {e}")

        # Terminate and join all processes
        for process in self.processes:
            process.join(timeout=5.0)
            if process.is_alive():
                try:
                    os.kill(process.pid, signal.SIGKILL)
                except OSError:
                    pass
                process.join(timeout=1.0)

        # Close all pipe connections
        for rank in range(self.dp_size):
            try:
                self.input_conns[rank].close()
            except OSError:
                pass
            try:
                self.output_conns[rank].close()
            except OSError:
                pass

        # Restore original pickle
        _disable_cloudpickle()


def update_vllm_config_for_dp_scheduler(vllm_config: Any) -> None:
    """
    Update vLLM configuration to use DPScheduler when DP size > 1.
    """
    dp_size = vllm_config.sharding_config.total_dp_size

    if dp_size > 1:
        if vllm_config.scheduler_config.async_scheduling:
            vllm_config.scheduler_config._original_scheduler_cls = AsyncScheduler
        else:
            vllm_config.scheduler_config._original_scheduler_cls = Scheduler

        vllm_config.scheduler_config.scheduler_cls = DPScheduler



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import hypothesis as hp
import hypothesis.strategies as hps
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized
from jax import random

import tpu_inference.kernels.structured_sparse_matmul.v1.spmm as structured_spmm


class StructuredSpmmTest(parameterized.TestCase):

    @parameterized.product(
        dtype=(jnp.float32, jnp.bfloat16, jnp.int8),
        stride=(1, 2, 3, 4),
        sparse_dim=(0, 1),
    )
    @hp.given(
        sparsity=hps.tuples(hps.integers(1, 3), hps.integers(4, 10)),
        shape_multiplier=hps.tuples(hps.integers(4, 10), hps.integers(4, 10)),
        seed=hps.integers(0, 4),
    )
    @hp.settings(deadline=None,
                 max_examples=20,
                 verbosity=hp.Verbosity.verbose)
    def test_sparse_matrix_encode(self, sparsity, stride, sparse_dim,
                                  shape_multiplier, dtype, seed):
        key = random.PRNGKey(seed)
        _, y = sparsity
        bitwidth = structured_spmm.next_pow2(structured_spmm.next_log2(y))
        packing = 32 // bitwidth
        n = shape_multiplier[0] * packing
        m = shape_multiplier[1]
        if sparse_dim == 0:
            n *= stride * y
        else:
            m *= stride * y
        shape = (n, m)
        mask = structured_spmm.gen_sparse_mask(key,
                                               shape,
                                               sparsity,
                                               sparse_dim=sparse_dim,
                                               stride=stride)
        data = random.normal(key, shape, dtype=jnp.float32).astype(dtype)
        sm = structured_spmm.Sparsifier(data,
                                        mask,
                                        sparsity=sparsity,
                                        sparse_dim=sparse_dim,
                                        stride=stride)
        decoded_data, decoded_mask = sm.decode()
        expected_data = jnp.where(mask, data, 0)
        np.testing.assert_array_equal(decoded_data, expected_data)
        np.testing.assert_array_equal(decoded_mask, mask)

    @parameterized.product(
        sparsity=((1, 4), (2, 5)),
        shape_spec=(
            (128, 128, 128),
            (512, 128, 256),
        ),
        rhs_sparse=(True, False),
        contract_sparse=(True, False),
        rhs_transpose=(True, False),
        default_val=(-1.9, 1.9),
        lhs_dtype=(jnp.float32, jnp.bfloat16),
        rhs_dtype=(jnp.float32, jnp.bfloat16),
    )
    def test_structured_spmm(
        self,
        sparsity,
        shape_spec,
        rhs_sparse,
        contract_sparse,
        rhs_transpose,
        default_val,
        lhs_dtype,
        rhs_dtype,
    ):
        # TODO(jevinjiang): support mix int/int matmtul in Mosaic.
        if lhs_dtype == rhs_dtype == jnp.int8:
            return
        _, y = sparsity
        m, k, n = shape_spec
        block_m = 128
        block_n = 128
        stride = 128
        if rhs_sparse:
            if contract_sparse:
                sparse_dim = int(rhs_transpose)
                k = math.lcm(stride * y, k)
            else:
                sparse_dim = int(not rhs_transpose)
                n = math.lcm(stride * y, n)
                block_n = math.lcm(stride * y, block_n)
        else:
            if contract_sparse:
                sparse_dim = 1
                k = math.lcm(stride * y, k)
            else:
                sparse_dim = 0
                m = math.lcm(stride * y, m)
                block_m = math.lcm(stride * y, block_m)
        lhs_shape = (m, k)
        rhs_shape = (n, k) if rhs_transpose else (k, n)

        key = random.PRNGKey(123)
        lhs = (random.normal(key, lhs_shape, dtype=jnp.float32) /
               100).astype(lhs_dtype)
        rhs = (random.normal(key, rhs_shape, dtype=jnp.float32) /
               100).astype(rhs_dtype)
        sparse_dtype = rhs_dtype if rhs_sparse else lhs_dtype
        default_val = (int(default_val) if jnp.issubdtype(
            sparse_dtype, jnp.integer) else default_val)
        out_dtype = structured_spmm._infer_out_dtype(lhs_dtype, rhs_dtype)

        if rhs_sparse:
            mask = structured_spmm.gen_sparse_mask(key,
                                                   rhs.shape,
                                                   sparsity,
                                                   sparse_dim=sparse_dim,
                                                   stride=stride)
            sm = structured_spmm.Sparsifier(rhs,
                                            mask,
                                            sparsity=sparsity,
                                            sparse_dim=sparse_dim,
                                            stride=stride)
            mat = lhs
            rhs = jnp.where(mask, rhs, default_val)
        else:
            mask = structured_spmm.gen_sparse_mask(key,
                                                   lhs.shape,
                                                   sparsity,
                                                   sparse_dim=sparse_dim,
                                                   stride=stride)
            sm = structured_spmm.Sparsifier(lhs,
                                            mask,
                                            sparsity=sparsity,
                                            sparse_dim=sparse_dim,
                                            stride=stride)
            mat = rhs
            lhs = jnp.where(mask, lhs, default_val)

        actual = structured_spmm.structured_spmm(
            sparsity,
            sm.nonzeros,
            sm.metadata,
            mat,
            rhs_sparse=rhs_sparse,
            contract_sparse=contract_sparse,
            rhs_transpose=rhs_transpose,
            stride=stride,
            block_m=block_m,
            block_k=k,
            block_n=block_n,
            default_value=default_val,
            out_dtype=out_dtype,
        )
        if rhs_transpose:
            rhs = rhs.T
        expected = jnp.dot(lhs, rhs, preferred_element_type=out_dtype)
        if out_dtype == jnp.bfloat16:
            actual = actual.astype(jnp.float32)
            expected = expected.astype(jnp.float32)
        np.testing.assert_allclose(actual, expected, rtol=5e-3, atol=5e-3)


if __name__ == "__main__":
    absltest.main()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import threading
from contextlib import contextmanager
from unittest.mock import MagicMock

import jax
import numpy as np
import pytest
from flax.typing import PRNGKey
from jax import numpy as jnp
from jax.sharding import Mesh
from vllm.config import ModelConfig
from vllm.model_executor.model_loader import LoadConfig, register_model_loader
from vllm.model_executor.model_loader.default_loader import DefaultModelLoader

from tpu_inference.layers.jax import JaxModule

logger = logging.getLogger(__name__)

GBYTES = 1024 * 1024 * 1024


@pytest.fixture
def rng() -> PRNGKey:
    """Provides a reusable JAX PRNGKey."""
    return jax.random.PRNGKey(42)


@pytest.fixture(scope="package")
def mesh():
    """
    Creates a mesh with 1 device.
    """
    if not jax.devices():
        pytest.skip("No JAX devices available for mesh creation.")

    devices = np.array(jax.local_devices()[:1])
    num_devices = len(devices)
    assert num_devices == 1
    device_mesh = devices.reshape((num_devices, 1, 1, 1))

    with Mesh(device_mesh,
              axis_names=('data', 'attn_dp', 'expert', 'model')) as m:
        yield m


@pytest.fixture
def mock_vllm_config():

    class MockVllmConfig:

        def __init__(self, model: str, kv_cache_dtype: str):
            self.model_config = ModelConfig(model)
            self.model_config.dtype = jnp.bfloat16
            self.load_config = LoadConfig(load_format="auto")
            self.load_config.download_dir = None
            self.cache_config = MagicMock(cache_dtype=kv_cache_dtype)
            self.quant_config = None
            self.additional_config = {}
            self.parallel_config = None

    return MockVllmConfig


@register_model_loader("skip_layers_model_loader_for_test")
class SkipLayersModelLoaderForTest(DefaultModelLoader):
    """Weight loader that skips layers beyond given limit.
    
    Some test are testing against weight loading, but it's meaningless
    to test all layers, assuming successfully loading the first few
    layers implies success of all layers. This special loader skips
    layers after given limit.
    """

    def __init__(self, load_config):
        self._num_layers_to_load = load_config.num_layers_to_load_for_test
        assert isinstance(self._num_layers_to_load, int)
        # `_prepare_weights` only recogonizes `load_format` from upstream.
        load_config.load_format = "auto"
        super().__init__(load_config)

    def get_all_weights(self, *args, **kwargs):
        for name, param in super().get_all_weights(*args, **kwargs):
            # If name matches "layers.\d+.", parse and skip if layer index is beyond limit
            match = re.search(r"layers\.(\d+)\.", name)
            if match:
                layer_idx = int(match.group(1))
                if layer_idx >= self._num_layers_to_load:
                    continue
            yield name, param


def count_model_param_bytes(model: JaxModule) -> int:
    """Count total bytes of all parameters in a model."""
    total = 0
    for _, param in model.named_parameters():
        if hasattr(param.get_value(), 'nbytes'):
            total += param.get_value().nbytes
    return total


class _MemoryPoller:
    """Background thread that polls device bytes_in_use and tracks the max.

    Polling is ~1μs per call on TPU, so even at sub-millisecond intervals
    the overhead is negligible relative to weight loading (seconds).
    """

    def __init__(self, device, poll_interval_s: float = 0.001):
        self._device = device
        self._poll_interval = poll_interval_s
        self._peak: int = 0
        self._num_samples: int = 0
        self._stop = threading.Event()
        self._thread = threading.Thread(target=self._poll_loop, daemon=True)

    def start(self, initial_bytes: int):
        self._peak = initial_bytes
        self._num_samples = 0
        self._stop.clear()
        self._thread.start()

    def stop(self) -> tuple[int, int]:
        """Stop polling and return (peak_bytes_observed, num_samples)."""
        self._stop.set()
        self._thread.join(timeout=5.0)
        return self._peak, self._num_samples

    def _poll_loop(self):
        while not self._stop.is_set():
            try:
                stats = self._device.memory_stats()
                if stats and "bytes_in_use" in stats:
                    current = stats["bytes_in_use"]
                    if current > self._peak:
                        self._peak = current
                    self._num_samples += 1
            except Exception:
                pass
            self._stop.wait(timeout=self._poll_interval)


def _get_bytes_in_use(device):
    """Read current bytes_in_use, or None if unavailable."""
    try:
        stats = device.memory_stats()
        return stats.get("bytes_in_use") if stats else None
    except Exception:
        return None


@pytest.fixture
def assert_weight_loading_memory_bounded():
    """Fixture that monitors device memory during weight loading.

    Spawns a background poller to capture peak bytes_in_use (window-scoped,
    not the cumulative high-water mark) and asserts peak delta stays within
    max(model_params * multiplier, min_threshold).

    Usage in tests::

        with assert_weight_loading_memory_bounded(model, "load_weights(...)"):
            loader.load_weights(model, model_config)
    """

    @contextmanager
    def _monitor(model,
                 description="operation",
                 threshold_multiplier=0.3,
                 min_threshold_bytes=2 * GBYTES):
        model_param_bytes = count_model_param_bytes(model)
        # Proportional threshold with absolute floor. PP-partitioned models
        # have fewer params but per-layer transients don't shrink
        # proportionally, so the floor prevents false failures.
        max_peak_delta = max(int(model_param_bytes * threshold_multiplier),
                             min_threshold_bytes)

        device = jax.local_devices()[0] if jax.local_devices() else None
        if device is None:
            logger.warning("No local devices; memory profiling skipped")
            yield
            return

        bytes_before = _get_bytes_in_use(device)
        if bytes_before is None:
            logger.warning("memory_stats() unavailable; profiling skipped")
            yield
            return

        poller = _MemoryPoller(device)
        poller.start(initial_bytes=bytes_before)

        yield

        peak_observed, num_samples = poller.stop()
        bytes_after = _get_bytes_in_use(device) or bytes_before
        peak_observed = max(peak_observed, bytes_after)
        peak_delta = peak_observed - bytes_before

        logger.info(
            "Memory profile [%s]: "
            "before=%.3f GB, after=%.3f GB, peak_observed=%.3f GB, "
            "peak_delta=%.3f GB, samples=%d, threshold=%.3f GB",
            description,
            bytes_before / GBYTES,
            bytes_after / GBYTES,
            peak_observed / GBYTES,
            peak_delta / GBYTES,
            num_samples,
            max_peak_delta / GBYTES,
        )

        assert peak_delta <= max_peak_delta, (
            f"Peak device memory during {description} exceeded threshold: "
            f"peak_delta={peak_delta / GBYTES:.3f} GB > "
            f"max={max_peak_delta / GBYTES:.3f} GB "
            f"(before={bytes_before / GBYTES:.3f} GB, "
            f"after={bytes_after / GBYTES:.3f} GB, "
            f"peak={peak_observed / GBYTES:.3f} GB, "
            f"{num_samples} samples)")

    return _monitor



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import numpy as jnp
from jax._src import test_util as jtu
from jax.sharding import Mesh

from tpu_inference.layers.jax.rope import (DeepseekScalingRotaryEmbedding,
                                           RotaryEmbedding)


class RotaryEmbeddingTest(jtu.JaxTestCase):

    def test_apply_rope(self):
        head_dim = 2
        rope_theta = 10000
        original_max_position_embeddings = 2
        rope = RotaryEmbedding(
            rotary_dim=head_dim,
            rope_theta=rope_theta,
            original_max_position_embeddings=original_max_position_embeddings,
            dtype=jnp.float32)
        rope.initialize_cache()
        self.assertTrue(
            rope.sin_cos_cache.shape == (original_max_position_embeddings,
                                         head_dim))
        expected_sin_cos = jnp.array([[1, 0], [0.5403023, 0.841471]],
                                     dtype=jnp.float32)
        self.assertArraysAllClose(rope.sin_cos_cache, expected_sin_cos)

        num_tokens = 2
        num_heads = 1
        positions = jnp.arange(num_tokens)
        x = jnp.ones((num_tokens, num_heads, head_dim))
        x_rope = rope.apply_rope(positions, x)
        expected_x_rope = jnp.array([[[1, 1]], [[-0.30116874, 1.3817732]]],
                                    dtype=jnp.float32)
        self.assertTrue(x_rope.shape == x.shape)
        self.assertArraysAllClose(x_rope, expected_x_rope)


class DeepseekScalingRotaryEmbeddingTest(jtu.JaxTestCase):

    def test_apply_rope(self):
        head_dim = 2
        rope_theta = 10000
        original_max_position_embeddings = 1
        scaling_factor = 2
        devices = jax.devices()
        mesh = Mesh(devices, ('data', ))

        rope = DeepseekScalingRotaryEmbedding(
            rotary_dim=head_dim,
            rope_theta=rope_theta,
            original_max_position_embeddings=original_max_position_embeddings,
            scaling_factor=scaling_factor,
            dtype=jnp.float32)
        with jax.set_mesh(mesh):
            rope.initialize_cache()
        expected_padded_dim = 128
        self.assertTrue(
            rope.sin_cos_cache.shape == (scaling_factor *
                                         original_max_position_embeddings,
                                         expected_padded_dim))

        valid_cache_slice = rope.sin_cos_cache[:, :head_dim]

        expected_sin_cos = jnp.array([[1.0693147, 0], [0.5777532, 0.8997973]],
                                     dtype=jnp.float32)

        self.assertArraysAllClose(valid_cache_slice, expected_sin_cos)

        num_tokens = 2
        num_heads = 1
        positions = jnp.arange(num_tokens)
        x = jnp.ones((num_tokens, num_heads, head_dim))
        x_rope = rope.apply_rope(positions, x)
        expected_x_rope = jnp.array(
            [[[1.0693147, 1.0693147]], [[-0.32204413, 1.4775505]]],
            dtype=jnp.float32)
        self.assertTrue(x_rope.shape == x.shape)
        self.assertArraysAllClose(x_rope, expected_x_rope)
